# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Set of HTML errors views. Error view are generally implemented
as startup views and are used for standard error pages (404, 500, etc.)
"""


from cubicweb_web.view import StartupView


class FourOhFour(StartupView):
    __regid__ = "404"

    def call(self):
        _ = self._cw._
        self.w("<h1>%s</h1>", _("this resource does not exist"))


class ErrorOccured(StartupView):
    __regid__ = "500"

    def call(self):
        _ = self._cw._
        self.w("<h1>%s</h1>", _("an error occurred, the request cannot be fulfilled"))



# copyright 2017-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Backward compatibility layer for CubicWeb to run as a Pyramid application."""

import os
import inspect
import itertools
import logging
import sys
import traceback
import warnings
from cgi import FieldStorage
from contextlib import contextmanager
from datetime import datetime
from urllib.parse import quote

import pyramid
from pyramid import httpexceptions
from pyramid import security
from pyramid import tweens
from pyramid.csrf import check_csrf_token, check_csrf_origin, get_csrf_token
from pyramid.httpexceptions import HTTPSeeOther, HTTPException
from pyramid.settings import asbool

import yams
import cubicweb
from cubicweb.debug import emit_to_debug_channel
from rql import BadRQLQuery

from cubicweb.devtools import BASE_URL

from cubicweb_web.application import CubicWebPublisher
from cubicweb_web.request import CubicWebRequestBase
from cubicweb_web.view import inject_html_generating_call_on_w
from cubicweb_web._exceptions import (
    PublishException,
    LogOut,
    Redirect,
    RequestError,
    RemoteCallFailed,
    NotFound,
)

log = logging.getLogger(__name__)


NO_SESSION_ERROR_MSG = (
    "No session factory registered, you can use pyramid_session_redis or "
    "include in pyramid.ini cubicweb.pyramid.session (which is not recommended "
    "for production instance. You can also see the Sessions chapter of the Pyramid documentation, "
    "if you want to define your own session factory."
)


class PyramidSessionHandler(object):
    """A CW Session handler that rely on the pyramid API to fetch the needed
    informations.

    It implements the :class:`cubicweb.web.application.CookieSessionHandler`
    API.
    """

    def __init__(self, appli):
        self.appli = appli

    def get_session(self, req):
        return req._request.cw_session

    def logout(self, req, goto_url):
        raise LogOut(url=goto_url)


def _cw_headers(request):
    return itertools.chain(
        *[
            [(k, item) for item in v]
            for k, v in request.cw_request.headers_out.getAllRawHeaders()
        ]
    )


@contextmanager
def cw_to_pyramid(request):
    """Context manager to wrap a call to the cubicweb API.

    All CW exceptions will be transformed into their pyramid equivalent.
    When needed, some CW reponse bits may be converted too (mainly headers)"""
    try:
        yield
    except Redirect as ex:
        assert 300 <= ex.status < 400
        raise httpexceptions.status_map[ex.status](
            ex.location, headers=_cw_headers(request)
        )
    except cubicweb.Unauthorized:
        raise httpexceptions.HTTPForbidden(
            request.cw_request._(
                "You're not authorized to access this page. "
                "If you think you should, please contact the site "
                "administrator."
            ),
            headers=_cw_headers(request),
        )
    except cubicweb.Forbidden:
        raise httpexceptions.HTTPForbidden(
            request.cw_request._(
                "This action is forbidden. "
                "If you think it should be allowed, please contact the site "
                "administrator."
            ),
            headers=_cw_headers(request),
        )
    except (BadRQLQuery, RequestError):
        raise


class CubicWebPyramidHandler(object):
    """A Pyramid request handler that rely on a cubicweb instance to do the
    whole job

    :param appli: A CubicWeb 'Application' object.
    """

    def __init__(self, appli, cubicweb_config):
        self.appli = appli

        if cubicweb_config["query-log-file"]:
            self._query_log = open(cubicweb_config["query-log-file"], "a")
            self._write_to_log = self._write_to_log_file
        else:
            self._write_to_log = self._write_to_logger

    def _write_to_log_file(self, text):
        self._query_log.write(text)
        self._query_log.flush()

    def _write_to_logger(self, text):
        log.info(text)

    def __call__(self, request):
        """
        Handler that mimics what CubicWebPublisher.main_handle_request and
        CubicWebPublisher.core_handle do
        """

        cubicweb_request = request.cw_request
        cubicweb_registry = request.registry["cubicweb.registry"]

        try:
            content = None
            try:
                with cw_to_pyramid(request):
                    controller_id, rset = self.appli.url_resolver.process(
                        cubicweb_request, cubicweb_request.path
                    )

                    try:
                        controller = cubicweb_registry["controllers"].select(
                            controller_id, cubicweb_request, appli=self.appli
                        )
                    except cubicweb.NoSelectableObject as ex:
                        warning_message = (
                            "failed to select a controller for this request "
                            f"{cubicweb_request.path} {request.method}."
                        )

                        if ex.objects:
                            candidates = "\n * ".join(repr(x) for x in ex.objects)
                            warning_message += (
                                " Here were the candidates controllers (but none matched): \n * "
                                f"{candidates}"
                            )
                        else:
                            warning_message += " There was no candidate controller."

                        log.warning(warning_message)

                        raise httpexceptions.HTTPBadRequest(
                            cubicweb_request._(
                                "couldn't handle this request as it is either badly formed or is "
                                "lacking the correct authorizations"
                            )
                        )

                    get_csrf_token(
                        request
                    )  # ensure that we have a CSRF token on all requests
                    safe_methods = frozenset(["GET", "HEAD", "OPTIONS", "TRACE"])
                    if request.method not in safe_methods and getattr(
                        controller, "require_csrf", True
                    ):
                        check_csrf_token(request)
                        check_csrf_origin(request)

                    self._write_to_log(
                        "REQUEST [%s] '%s' selected controller %s at %s:%s"
                        % (
                            controller_id,
                            cubicweb_request.path,
                            controller,
                            inspect.getsourcefile(controller.__class__),
                            inspect.getsourcelines(controller.__class__)[1],
                        )
                    )
                    emit_to_debug_channel(
                        "vreg",
                        {
                            "vreg": cubicweb_registry,
                        },
                    )
                    emit_to_debug_channel(
                        "controller",
                        {
                            "kind": controller_id,
                            "request": cubicweb_request,
                            "path": cubicweb_request.path,
                            "controller": controller,
                            "config": self.appli.repo.config,
                        },
                    )

                    cubicweb_request.update_search_state()
                    content = controller.publish(rset=rset)

                    # XXX this auto-commit should be handled by the cw_request
                    # cleanup or the pyramid transaction manager.
                    # It is kept here to have the ValidationError handling bw
                    # compatible
                    if cubicweb_request.cnx:
                        transaction_uuid = cubicweb_request.cnx.commit()
                        # commited = True
                        if transaction_uuid is not None:
                            cubicweb_request.data[
                                "last_undoable_transaction"
                            ] = transaction_uuid
            except yams.ValidationError as ex:
                # XXX The validation_error_handler implementation is light, we
                # should redo it better in cw_to_pyramid, so it can be properly
                # handled when raised from a cubicweb view.
                # BUT the real handling of validation errors should be done
                # earlier in the controllers, not here. In the end, the
                # ValidationError should never by handled here.
                content = self.appli.validation_error_handler(cubicweb_request, ex)
            except RemoteCallFailed as exception:
                raise pyramid.httpexceptions.exception_response(
                    exception.status,
                    body=exception.dumps(),
                    content_type="application/json",
                    charset="utf-8",
                )

            if content is not None:
                request.response.body = content

        except LogOut as ex:
            # The actual 'logging out' logic should be in separated function
            # that is accessible by the pyramid views
            headers = security.forget(request)
            raise HTTPSeeOther(ex.url, headers=headers)
        except cubicweb.AuthenticationError:
            # Will occur upon access to cubicweb_request.cnx which is a
            # cubicweb.dbapi._NeedAuthAccessMock.
            if not content:
                params = ""

                new_path = request.path_qs
                if new_path != "/":
                    params = f"?postlogin_path={quote(new_path)}"

                raise HTTPSeeOther(f"/login{params}")
        except NotFound as ex:
            if not cubicweb_request.cnx:
                new_path = request.path_qs
                params = f"?postlogin_path={quote(new_path)}" if new_path != "/" else ""
                raise HTTPSeeOther(f"/login{params}")
            view = cubicweb_registry["views"].select("404", cubicweb_request)
            content = cubicweb_registry["views"].main_template(
                cubicweb_request, view=view
            )
            request.response.status_code = ex.status
            request.response.body = content
        finally:
            # XXX CubicWebPyramidRequest.headers_out should
            # access directly the pyramid response headers.
            request.response.headers.clear()
            for (
                header_name,
                header_values,
            ) in cubicweb_request.headers_out.getAllRawHeaders():
                for item in header_values:
                    request.response.headers.add(header_name, item)

        return request.response

    def error_handler(self, exc, request):
        req = request.cw_request
        if isinstance(exc, httpexceptions.HTTPException):
            request.response = exc
        elif isinstance(exc, PublishException) and exc.status is not None:
            request.response = httpexceptions.exception_response(exc.status)
        else:
            request.response = httpexceptions.HTTPInternalServerError()
        request.response.cache_control = "no-cache"
        vreg = request.registry["cubicweb.registry"]
        excinfo = sys.exc_info()
        req.reset_message()
        if req.ajax_request:
            content = self.appli.ajax_error_handler(req, exc)
        else:
            try:
                req.data["ex"] = exc
                req.data["excinfo"] = excinfo
                errview = vreg["views"].select("error", req)
                template = self.appli.main_template_id(req)
                content = vreg["views"].main_template(req, template, view=errview)
            except Exception:
                content = vreg["views"].main_template(req, "error-template")
        log.exception(exc)
        request.response.body = content
        return request.response


class CubicWebPyramidRequest(CubicWebRequestBase):
    """A CubicWeb request that only wraps a pyramid request.

    :param request: A pyramid request

    """

    def __init__(self, request):
        self._request = request
        vreg = request.registry["cubicweb.registry"]

        base_url = vreg.config["base-url"]

        if "PYTEST_CURRENT_TEST" in os.environ:
            assert request.url.startswith(base_url), (
                f"Error: request {request} must start with the BASE_URL {BASE_URL} when running "
                "tests. Please prefix it with cubicweb.devtools.BASE_URL"
            )

        self.path = request.upath_info

        post = request.params.mixed()
        headers_in = request.headers

        super().__init__(vreg, post, headers=headers_in)

        self.content = request.body_file_seekable

    def __repr__(self):
        """
        Will render something like:

        <cubicweb_web.bwcompat.CubicWebPyramidRequest object at 0x.... GET https://example.com/path/>
        """
        return f"{super().__repr__()[:-1]} {self._request.method} {self._request.url}>"

    def setup_params(self, params):
        self._uncleaned_form = {}
        for param, val in params.items():
            if param in self.no_script_form_params and val:
                val = self.no_script_form_param(param, val)
            if isinstance(val, FieldStorage) and val.file:
                val = (val.filename, val.file)
            if param == "_cwmsgid":
                self.set_message_id(val)
            else:
                self._uncleaned_form[param] = val

    @property
    def form(self):
        class_name = self.__class__.__name__
        warnings.warn(
            f"{class_name}.form is deprecated and will be remove, uses "
            f"{class_name}.get_cleaned_form using a form_validator instead"
        )
        return self._uncleaned_form

    @form.setter
    def form(self, new_form):
        class_name = self.__class__.__name__
        warnings.warn(
            f"{class_name}.form is deprecated and will be remove, you won't be able to set it "
            "explictely, better options for specific cases where modifying it is needed"
        )
        self._uncleaned_form = new_form

    def get_cleaned_form(self, form_validator):
        return form_validator.validate(self._uncleaned_form)

    def relative_path(self, includeparams=True):
        path = self._request.path_info[1:]
        if includeparams and self._request.query_string:
            return f"{path}?{self._request.query_string}"
        return path

    def instance_uri(self):
        return self._request.application_url

    def get_full_path(self):
        path = self._request.path
        if self._request.query_string:
            return f"{path}?{self._request.query_string}"
        return path

    def http_method(self):
        return self._request.method

    def _set_status_out(self, value):
        self._request.response.status_int = value

    def _get_status_out(self):
        return self._request.response.status_int

    status_out = property(_get_status_out, _set_status_out)

    @property
    def message(self):
        """Returns a '<br>' joined list of the cubicweb current message and the
        default pyramid flash queue messages.
        """
        try:
            return "\n<br>\n".join(
                self._request.session.pop_flash()
                + self._request.session.pop_flash("cubicweb")
            )
        except AttributeError:
            log.exception(NO_SESSION_ERROR_MSG)
            return ""

    def set_message(self, msg):
        try:
            self.reset_message()
            self._request.session.flash(msg, "cubicweb")
        except AttributeError:
            log.exception(NO_SESSION_ERROR_MSG)

    def set_message_id(self, msgid):
        try:
            self.reset_message()
            self.set_message(self._request.session.pop(msgid, ""))
        except AttributeError:
            log.exception(NO_SESSION_ERROR_MSG)

    def reset_message(self):
        try:
            self._request.session.pop_flash("cubicweb")
        except AttributeError:
            log.exception(NO_SESSION_ERROR_MSG)


def _cw_request(request):
    """Obtains a CubicWeb request wrapper for the pyramid request.

    :param request: A pyramid request
    :return: A CubicWeb request
    :returns type: :class:`CubicWebPyramidRequest`

    Not meant for direct use, use ``request.cw_request`` instead.

    """
    req = CubicWebPyramidRequest(request)
    cnx = request.cw_cnx
    if cnx is not None:
        req.set_cnx(request.cw_cnx)
    return req


class TweenHandler(object):
    """A Pyramid tween handler that submit unhandled requests to a Cubicweb
    handler.

    The CubicWeb handler to use is expected to be in the pyramid registry, at
    key ``'cubicweb.handler'``.
    """

    def __init__(self, handler, registry):
        self.handler = handler
        self.cwhandler = registry["cubicweb.handler"]

    def __call__(self, request):
        view_kind = "pyramid"
        now = str(datetime.now()).split(".")[0]

        try:
            try:
                response = self.handler(request)
            except httpexceptions.HTTPNotFound:
                view_kind = "cubicweb"
                response = self.cwhandler(request)

            print(
                f"{now} - ({view_kind} view) "
                f'"{request.method} {request.path}'
                f' {request.http_version}" '
                f"{response.status_code} {len(response.body)}"
            )

        except HTTPException as e:
            print(
                f'{now} - ({view_kind} view) "{request.method} '
                f'{request.path} {request.http_version}" '
                f"{e.code} {len(e.body)}"
            )

            # we don't want a traceback for a redirection, only for errors or others
            if not (300 <= e.code < 400):
                traceback.print_exc()
            raise

        return response


def render_view(request, vid, **kwargs):
    """Helper function to render a CubicWeb view.

    :param request: A pyramid request
    :param vid: A CubicWeb view id
    :param kwargs: Keyword arguments to select and instanciate the view
    :returns: The rendered view content
    """
    vreg = request.registry["cubicweb.registry"]
    # XXX The select() function could, know how to handle a pyramid
    # request, and feed it directly to the views that supports it.
    # On the other hand, we could refine the View concept and decide it works
    # with a cnx, and never with a WebRequest

    with cw_to_pyramid(request):
        view = vreg["views"].select(vid, request.cw_request, **kwargs)
        view.set_stream()
        view.render()
        return view._stream.getvalue()


def includeme(config):
    """Set up a tween app that will handle the request if the main application
    raises a HTTPNotFound exception.

    This is to keep legacy compatibility for cubes that makes use of the
    cubicweb urlresolvers.

    It provides, for now, support for cubicweb controllers, but this feature
    will be reimplemented separatly in a less compatible way.

    It is automatically included by the configuration system, but can be
    disabled in the :ref:`pyramid_settings`:

    .. code-block:: ini

        cubicweb.bwcompat = no
    """

    cwconfig = config.registry["cubicweb.config"]
    repository = config.registry["cubicweb.repository"]

    config.add_request_method(_cw_request, name="cw_request", property=True, reify=True)

    cwappli = CubicWebPublisher(
        repository, cwconfig, session_handler_fact=PyramidSessionHandler
    )
    cwhandler = CubicWebPyramidHandler(cwappli, cwconfig)

    config.registry["cubicweb.appli"] = cwappli
    config.registry["cubicweb.handler"] = cwhandler

    config.add_tween("cubicweb_web.bwcompat.TweenHandler", under=tweens.EXCVIEW)

    if asbool(config.registry.settings.get("cubicweb.bwcompat.errorhandler", True)):
        config.add_view(cwhandler.error_handler, context=Exception)
        # XXX why do i need this?
        config.add_view(cwhandler.error_handler, context=httpexceptions.HTTPForbidden)

    if cwconfig.in_debug_mode:
        # this is for injecting those into generated html:
        # > cubicweb-generated-by="module.Class" cubicweb-from-source="/path/to/file.py:42"
        inject_html_generating_call_on_w()



# copyright 2017-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

""" Provide login views that reproduce a classical CubicWeb behavior"""
import logging
import warnings

from urllib.parse import urlparse

from pyramid import security
from pyramid.httpexceptions import HTTPSeeOther, HTTPNotFound
from pyramid.settings import asbool
from pyramid.view import view_config

import cubicweb
from cubicweb_web.utils import url_path_starts_with_prefix, remove_prefix_from_url_path
from cubicweb_web.bwcompat import render_view

logger = logging.getLogger(__name__)


@view_config(route_name="login")
def login_form(request):
    """Default view for the 'login' route.

    Display the 'login' CubicWeb view, which is should be a login form"""
    request.response.text = render_view(request, "login")
    return request.response


@view_config(
    route_name="login", request_param=("__login", "__password"), request_method="POST"
)
def login_password_login(request):
    """Handle POST of __login/__password on the 'login' route.

    The authentication itself is delegated to the CubicWeb repository.

    Request parameters:

    :param __login: The user login (or email if :confval:`allow-email-login` is
                    on.
    :param __password: The user password
    :param __setauthcookie: (optional) If defined and equal to '1', set the
                            authentication cookie maxage to 1 week.

                            If not, the authentication cookie is a session
                            cookie.
    """
    repo = request.registry["cubicweb.repository"]

    login = request.params["__login"]
    password = request.params["__password"]

    try:
        with repo.internal_cnx() as cnx:
            user = repo.authenticate_user(cnx, login, password=password)
            user_eid = user.eid
    except cubicweb.AuthenticationError:
        if repo.vreg.config.get("language-mode") != "":
            lang = request.cw_request.negotiated_language()
            if lang is not None:
                request.cw_request.set_language(lang)
        request.cw_request.set_message(
            request.cw_request._(
                "Authentication failed. Please check your credentials."
            )
        )
        request.cw_request.post = dict(request.params)
        del request.cw_request.post["__password"]
        request.response.status_code = 403
        return login_form(request)

    headers = security.remember(
        request,
        user_eid,
        persistent=asbool(request.params.get("__setauthcookie", False)),
    )

    new_path = request.params.get("postlogin_path", "")
    base_url = repo.vreg.config["base-url"]
    route_prefix = urlparse(base_url).path.lstrip("/")

    if repo.vreg.config["receives-base-url-path"] and route_prefix:
        if not url_path_starts_with_prefix(new_path, route_prefix):
            logger.warning(
                f"Error 404: login view tries to redirect to the path '{new_path}' but it doesn't "
                f"start with the route_prefix '{route_prefix}'"
            )
            raise HTTPNotFound()
        new_path = remove_prefix_from_url_path(new_path, route_prefix)
    else:
        if route_prefix:
            warnings.warn(
                "in your configuration your base url "
                f"'{repo.vreg.config['base-url']}' has a route prefix: "
                f"'{route_prefix}' but the option 'receives-base-url-path' is set to False.\n\n"
                "You should set it to True and avoid relying on nginx/apache for url rewritting."
            )

        new_path = new_path.lstrip("/")

    if new_path == "login":
        new_path = ""

    url = request.cw_request.build_url(new_path)
    raise HTTPSeeOther(url, headers=headers)


@view_config(route_name="login", effective_principals=security.Authenticated)
def login_already_loggedin(request):
    """'login' route view for Authenticated users.

    Simply redirect the user to '/'."""
    raise HTTPSeeOther(request.params.get("postlogin_path", "/"))


def includeme(config):
    """Create the 'login' route ('/login') and load this module views"""
    cwconfig = config.registry["cubicweb.config"]
    config.add_route("login", "/login")
    if cwconfig.get("language-mode") == "url-prefix":
        config.add_route("login-lang", "/{lang}/login")
        config.add_view(
            login_already_loggedin,
            route_name="login-lang",
            effective_principals=security.Authenticated,
        )
        config.add_view(login_form, route_name="login-lang")
        config.add_view(
            login_password_login,
            route_name="login-lang",
            request_param=("__login", "__password"),
            request_method="POST",
        )
    config.scan(".login")



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Associate url's path to view identifier / rql queries.

CubicWeb finds all registered URLPathEvaluators, orders them according
to their ``priority`` attribute and calls their ``evaluate_path()``
method. The first that returns something and doesn't raise a
``PathDontMatch`` exception wins.

Here is the default evaluator chain:

1. :class:`cubicweb_web.views.urlpublishing.RawPathEvaluator` handles
   unique url segments that match exactly one of the registered
   controller's *__regid__*. Urls such as */view?*, */edit?*, */json?*
   fall in that category;

2. :class:`cubicweb_web.views.urlpublishing.EidPathEvaluator` handles
   unique url segments that are eids (e.g. */1234*);

3. :class:`cubicweb_web.views.urlpublishing.URLRewriteEvaluator`
   selects all urlrewriter components, sorts them according to their
   priority, call their ``rewrite()`` method, the first one that
   doesn't raise a ``KeyError`` wins. This is where the
   :mod:`cubicweb_web.views.urlrewrite` and
   :class:`cubicweb_web.views.urlrewrite.SimpleReqRewriter` comes into
   play;

4. :class:`cubicweb_web.views.urlpublishing.RestPathEvaluator` handles
   urls based on entity types and attributes : <etype>((/<attribute
   name>])?/<attribute value>)?  This is why ``cwuser/carlos`` works;

5. :class:`cubicweb_web.views.urlpublishing.ActionPathEvaluator`
   handles any of the previous paths with an additional trailing
   "/<action>" segment, <action> being one of the registered actions'
   __regid__.


.. note::

 Actionpath executes a query whose results is lost
 because of redirecting instead of direct traversal.
"""

import logging
import warnings

from urllib.parse import urlparse

from rql import TypeResolverException

from cubicweb import RegistryException
from cubicweb_web import NotFound, Redirect, component, views
from cubicweb_web.utils import url_path_starts_with_prefix, remove_prefix_from_url_path

logger = logging.getLogger(__name__)


class PathDontMatch(Exception):
    """exception used by url evaluators to notify they can't evaluate
    a path
    """


class URLPublisherComponent(component.Component):
    """Associate url path to view identifier / rql queries, by
    applying a chain of urlpathevaluator components.

    An evaluator is a URLPathEvaluator subclass with an .evaluate_path
    method taking the request object and the path to publish as
    argument.  It will either return a publishing method identifier
    and an rql query on success or raise a `PathDontMatch` exception
    on failure. URL evaluators are called according to their
    `priority` attribute, with 0 as the greatest priority and greater
    values as lower priority. The first evaluator returning a result
    or raising something else than `PathDontMatch` will stop the
    handlers chain.
    """

    __regid__ = "urlpublisher"
    vreg = None  # XXX necessary until property for deprecation warning is on appobject

    def __init__(self, vreg, default_method="view"):
        super().__init__()
        self.vreg = vreg
        self.default_method = default_method
        evaluators = []
        for evaluatorcls in vreg["components"]["urlpathevaluator"]:
            # instantiation needed
            evaluator = evaluatorcls(self)
            evaluators.append(evaluator)
        self.evaluators = sorted(evaluators, key=lambda x: x.priority)

    def process(self, req, path):
        """Given a URL (essentially characterized by a path on the
        server, but additional information may be found in the request
        object), return a publishing method identifier
        (e.g. controller) and an optional result set.

        :type req: `cubicweb_web.request.CubicWebRequestBase`
        :param req: the request object

        :type path: str
        :param path: the path of the resource to publish. If empty, None or "/"
                     "view" is used as the default path.

        :rtype: tuple(str, `cubicweb.rset.ResultSet` or None)
        :return: the publishing method identifier and an optional result set

        :raise NotFound: if no handler is able to decode the given path
        """

        base_url = self.vreg.config["base-url"]
        route_prefix = urlparse(base_url).path.lstrip("/")

        datadir_url = self.vreg.config["datadir-url"]
        datadir_route_prefix = urlparse(datadir_url).path if datadir_url else None

        if self.vreg.config["receives-base-url-path"] and (
            route_prefix or datadir_route_prefix
        ):
            # we are not dealing with a data url configured on datadir-url
            # case of custom datadir-url is handle in DataController
            if not (
                datadir_url and url_path_starts_with_prefix(path, datadir_route_prefix)
            ):
                if not url_path_starts_with_prefix(path, route_prefix):
                    logger.warning(
                        f"Error 404: instance has recieved the request '{req}' with path '{path}' but "
                        f"it doesn't start with the route_prefix '{route_prefix}'"
                    )
                    raise NotFound()

                path = remove_prefix_from_url_path(path, route_prefix)

        elif route_prefix:
            warnings.warn(
                "in your configuration your base url "
                f"'{self.vreg.config['base-url']}' has a route prefix: "
                f"'{route_prefix}' but the option 'receives-base-url-path' is set to False.\n\n"
                "You should set it to True and avoid relying on nginx/apache for url rewritting."
            )

        parts = [part for part in path.split("/") if part != ""] or [
            self.default_method
        ]

        if not [part for part in path.split("/") if part != ""]:
            logger.debug(
                f"couldn't extract parts from url path {path}, use default method "
                f"{self.default_method} instead"
            )

        language_mode = self.vreg.config.get("language-mode")
        if (
            language_mode == "url-prefix"
            and parts
            and parts[0] in self.vreg.config.available_languages()
        ):
            # language from URL
            req.set_language(parts.pop(0))
            path = "/".join(parts)
            # if parts only contains lang, use 'view' default path
            if not parts:
                parts = (self.default_method,)
        elif language_mode in ("http-negotiation", "url-prefix"):
            # negotiated language
            lang = req.negotiated_language()
            if lang:
                if language_mode == "http-negotiation":
                    req.set_language_from_req(lang)
                else:
                    req.set_language(lang)
        origin_rql = req.form.get("rql")
        if origin_rql:
            if parts[0] in self.vreg["controllers"]:
                return parts[0], None
            return "view", None
        for evaluator in self.evaluators:
            try:
                logger.debug(
                    f"try evaluator '{evaluator}' on request {req} with '{parts}'"
                )
                pmid, rset = evaluator.evaluate_path(req, parts[:])
                break
            except PathDontMatch:
                continue
        else:
            logger.debug(f"failed to get an evaluator for url {req}, raise 404")
            raise NotFound(path)

        logger.debug(f"evaluator {evaluator} matched on request {req}")

        # Store if the final rql (changed by evaluator.evaluate_path) comes from
        # the end user (?rql=Any X Where is CWUser) or comes from a rewrite rule
        # such as SimpleReqRewriter.
        new_rql = req.form.get("rql")
        req._rql_rewritten = origin_rql != new_rql

        if pmid is None:
            pmid = self.default_method
        return pmid, rset


class URLPathEvaluator(component.Component):
    __abstract__ = True
    __regid__ = "urlpathevaluator"
    vreg = None  # XXX necessary until property for deprecation warning is on appobject

    def __init__(self, urlpublisher):
        self.urlpublisher = urlpublisher
        self.vreg = urlpublisher.vreg


class RawPathEvaluator(URLPathEvaluator):
    """handle path of the form::

    <publishing_method>?parameters...
    """

    priority = 0

    def evaluate_path(self, req, parts):
        if len(parts) == 1 and parts[0] in self.vreg["controllers"]:
            return parts[0], None
        logging.debug(
            f"{self} doesn't match request {req} because parts {parts} is not of length one "
            f"or first element of part is not in available controllers: {self.vreg['controllers']}"
        )
        raise PathDontMatch()


class EidPathEvaluator(URLPathEvaluator):
    """handle path with the form::

    <eid>
    """

    priority = 1

    def evaluate_path(self, req, parts):
        if len(parts) != 1:
            logging.debug(
                f"{self} doesn't match request {req} because parts {parts} is not of length one"
            )
            raise PathDontMatch()
        try:
            rset = req.execute("Any X WHERE X eid %(x)s", {"x": int(parts[0])})
        except ValueError as e:
            logging.debug(
                f"{self} doesn't match request {req} because exception {e} got raised while trying "
                f"to convert '{parts[0]}' into an int then runt the query "
                f"'Any X WHERE X eid {parts[0]}'"
            )
            raise PathDontMatch()
        if rset.rowcount == 0:
            logging.debug(
                f"{self} query 'Any X WHERE X eid {parts[0]}' returned nothing, raise 404"
            )
            raise NotFound()
        return None, rset


class RestPathEvaluator(URLPathEvaluator):
    """handle path with the form::

    <etype>[[/<attribute name>]/<attribute value>]*
    """

    priority = 3

    def evaluate_path(self, req, parts):
        if not (0 < len(parts) < 4):
            logging.debug(
                f"{self} doesn't match request {req} because parts not (0 < len({parts}) < 4)"
            )
            raise PathDontMatch()
        try:
            _debug_first_part = parts[0]
            etype = self.vreg.case_insensitive_etypes[parts.pop(0).lower()]
        except KeyError:
            logging.debug(
                f"{self} doesn't match request {req} because its first part "
                f"{_debug_first_part.lower()} is not in available etypes: "
                f"{self.vreg.case_insensitive_etypes}"
            )
            raise PathDontMatch()
        cls = self.vreg["etypes"].etype_class(etype)
        if parts:
            if len(parts) == 2:
                attrname = parts.pop(0).lower()
                try:
                    cls.e_schema.subject_relations[attrname]
                except KeyError:
                    logging.debug(
                        f"{self} doesn't match request {req} because its second part {attrname} "
                        f"is not in available subject_relations {cls.e_schema.subject_relations}"
                    )
                    raise PathDontMatch()
            else:
                attrname = cls.cw_rest_attr_info()[0]
            value = req.url_unquote(parts.pop(0))
            return self.handle_etype_attr(req, cls, attrname, value)
        return self.handle_etype(req, cls)

    def set_vid_for_rset(self, req, cls, rset):  # cls is there to ease overriding
        if rset.rowcount == 0:
            raise NotFound()
        if "vid" not in req.form:
            # check_table=False tells vid_from_rset not to try to use a table view if fetch_rql
            # include some non final relation
            req.form["vid"] = views.vid_from_rset(
                req, rset, req.vreg.schema, check_table=False
            )

    def handle_etype(self, req, cls):
        rset = req.execute(cls.fetch_rql(req.user))
        self.set_vid_for_rset(req, cls, rset)
        return None, rset

    def handle_etype_attr(self, req, cls, attrname, value):
        st = cls.fetch_rqlst(req.user, ordermethod=None)
        st.add_constant_restriction(st.get_variable("X"), attrname, "x", "Substitute")
        if attrname == "eid":
            try:
                rset = req.execute(st.as_string(), {"x": int(value)})
            except (ValueError, TypeResolverException):
                logging.debug(
                    f"{self} doesn't match request {req} because the query '{st.as_string()}' with "
                    f"'x = {value}' failed"
                )
                # conflicting eid/type
                raise PathDontMatch()
        else:
            rset = req.execute(st.as_string(), {"x": value})
        self.set_vid_for_rset(req, cls, rset)
        return None, rset


class URLRewriteEvaluator(URLPathEvaluator):
    """tries to find a rewrite rule to apply

    URL rewrite rule definitions are stored in URLRewriter objects
    """

    priority = 2

    def evaluate_path(self, req, parts):
        # uri <=> req._twreq.path or req._twreq.uri
        uri = req.url_unquote("/" + "/".join(parts))
        evaluators = sorted(
            self.vreg["urlrewriting"].all_objects(),
            key=lambda x: x.priority,
            reverse=True,
        )

        for rewritercls in evaluators:
            rewriter = rewritercls(req)
            try:
                logger.debug(f"{self} try to rewrite url using {rewriter}")

                # XXX we might want to chain url rewrites
                return rewriter.rewrite(req, uri)
            except KeyError:
                logger.debug(f"{self} Error: {rewriter} failed to rewrite url")
                continue
        logger.debug(f"{self} couldn't find a rewriter to rewrite url")
        raise PathDontMatch()


class ActionPathEvaluator(URLPathEvaluator):
    """handle path with the form::

    <any evaluator path>/<action>
    """

    priority = 4

    def evaluate_path(self, req, parts):
        if len(parts) < 2:
            logging.debug(
                f"{self} doesn't match request {req} because parts is len(parts) < 2"
            )
            raise PathDontMatch()
        # remove last part and see if this is something like an actions
        # if so, call
        # XXX bad smell: refactor to simpler code
        try:
            actionsreg = self.vreg["actions"]
            requested = parts.pop(-1)
            actions = actionsreg[requested]
        except RegistryException:
            logging.debug(
                f"{self} doesn't match request {req} because parts is len(parts) < 2"
            )
            raise PathDontMatch()
        for evaluator in self.urlpublisher.evaluators:
            logging.debug(f"{self} try to evaluate url using evaluator {evaluator}")

            if evaluator is self or evaluator.priority == 0:
                continue
            try:
                pmid, rset = evaluator.evaluate_path(req, parts[:])
            except PathDontMatch:
                continue
            else:
                try:
                    action = actionsreg._select_best(actions, req, rset=rset)
                    if action is not None:
                        logging.debug(
                            f"{self} evaluator {evaluator} matched, redirect to {action.url()}"
                        )
                        raise Redirect(action.url())
                except RegistryException:
                    pass  # continue searching

        logging.debug(f"{self} couldn't find any evaluator for request {req}")
        raise PathDontMatch()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb_web import facet
from cubicweb_web.devtools.testlib import WebCWTC


class InsertAttrRelationCWTC(WebCWTC):
    def parse(self, query):
        with self.admin_access.cnx() as cnx:
            rqlst = self.vreg.parse(cnx, query)
            rqlst.children[0]
        return rqlst

    def _generate(self, rqlst, rel, role, attr):
        select = rqlst.children[0]
        filtered_variable = facet.get_filtered_variable(select)
        facet.prepare_select(select, filtered_variable)
        facet.insert_attr_select_relation(select, filtered_variable, rel, role, attr)
        return rqlst.as_string()

    @property
    def select(self):
        return self.parse(
            "Any B,(NOW - CD),S,V,U,GROUP_CONCAT(TN),VN,P,CD,BMD "
            "GROUPBY B,CD,S,V,U,VN,P,BMD "
            "WHERE B in_state S, B creation_date CD, "
            "B modification_date BMD, T? tags B, T name TN, "
            "V? bookmarked_by B, V title VN, B created_by U?, "
            'B in_group P, P name "managers"'
        )

    def test_1(self):
        self.assertEqual(
            self._generate(self.select, "in_state", "subject", "name"),
            'DISTINCT Any A,C ORDERBY C WHERE B in_group P, P name "managers", '
            "B in_state A, B is CWUser, A name C",
        )

    def test_2(self):
        self.assertEqual(
            self._generate(self.select, "tags", "object", "name"),
            'DISTINCT Any A,C ORDERBY C WHERE B in_group P, P name "managers", '
            "A tags B, B is CWUser, A name C",
        )

    def test_3(self):
        self.assertEqual(
            self._generate(self.select, "created_by", "subject", "login"),
            'DISTINCT Any A,C ORDERBY C WHERE B in_group P, P name "managers", '
            "B created_by A, B is CWUser, A login C",
        )

    def test_4(self):
        self.assertEqual(
            self._generate(
                self.parse("Any X WHERE X is CWUser"), "created_by", "subject", "login"
            ),
            "DISTINCT Any A,B ORDERBY B WHERE X is CWUser, X created_by A, A login B",
        )

    def test_5(self):
        self.assertEqual(
            self._generate(
                self.parse("Any X,L WHERE X is CWUser, X login L"),
                "created_by",
                "subject",
                "login",
            ),
            "DISTINCT Any A,B ORDERBY B WHERE X is CWUser, X created_by A, A login B",
        )

    def test_nonregr1(self):
        select = self.parse(
            "Any T,V WHERE T bookmarked_by V?, "
            'V in_state VS, VS name "published", T created_by U'
        )
        self.assertEqual(
            self._generate(select, "created_by", "subject", "login"),
            "DISTINCT Any A,B ORDERBY B WHERE T created_by U, "
            "T created_by A, T is Bookmark, A login B",
        )

    def test_nonregr2(self):
        select = self.parse(
            "DISTINCT Any V,TN,L ORDERBY TN,L WHERE T nom TN, V connait T, T is Personne, V is CWUser,"
            'NOT V in_state VS, VS name "published", V login L'
        )
        rschema = self.schema["connait"]
        for rdefs in rschema.relation_definitions.values():
            rdefs.cardinality = "++"
        try:
            self.assertEqual(
                self._generate(select, "in_state", "subject", "name"),
                "DISTINCT Any A,B ORDERBY B WHERE V is CWUser, "
                'NOT EXISTS(V in_state VS), VS name "published", '
                "V in_state A, A name B",
            )
        finally:
            for rdefs in rschema.relation_definitions.values():
                rdefs.cardinality = "**"

    def test_nonregr3(self):
        select = self.parse(
            "DISTINCT Any X, MAX(Y) GROUPBY X WHERE X is CWUser, Y is Bookmark, X in_group A"
        )
        self.assertEqual(
            self._generate(select, "in_group", "subject", "name"),
            "DISTINCT Any B,C ORDERBY C WHERE X is CWUser, X in_group B, B name C",
        )


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""CubicWeb web client core. You'll need a apache-modpython or twisted
publisher to get a full CubicWeb web application
"""

from logging import getLogger

from logilab.common.deprecation import callable_moved

from cubicweb import _
from cubicweb._exceptions import (
    Forbidden,
    Unauthorized,
    ValidationError,
)
from cubicweb.utils import json_dumps
from pyramid.config import Configurator

from cubicweb_web._exceptions import (
    DirectResponse,
    InvalidSession,
    LogOut,
    NotFound,
    NothingToEdit,
    ProcessFormError,
    PublishException,
    Redirect,
    RemoteCallFailed,
    RequestError,
    StatusResponse,
)

assert json_dumps is not None, "no json module installed"

json_dumps = callable_moved("cubicweb.utils", "json_dumps")
eid_param = callable_moved("cubicweb.uilib", "eid_param")

INTERNAL_FIELD_VALUE = "__cubicweb_internal_field__"


class stdmsgs:
    """standard ui message (in a class for bw compat)"""

    BUTTON_OK = (_("button_ok"), "OK_ICON")
    BUTTON_APPLY = (_("button_apply"), "APPLY_ICON")
    BUTTON_CANCEL = (_("button_cancel"), "CANCEL_ICON")
    BUTTON_DELETE = (_("button_delete"), "TRASH_ICON")
    YES = (_("yes"), None)
    NO = (_("no"), None)


LOGGER = getLogger("cubicweb_web")

# XXX deprecated
FACETTES = set()


def jsonize(function):
    def newfunc(*args, **kwargs):
        value = function(*args, **kwargs)
        try:
            return json_dumps(value)
        except TypeError:
            return json_dumps(repr(value))

    return newfunc


__all__ = [
    "DirectResponse",
    "Forbidden",
    "InvalidSession",
    "LogOut",
    "NotFound",
    "NothingToEdit",
    "ProcessFormError",
    "PublishException",
    "Redirect",
    "RemoteCallFailed",
    "RequestError",
    "StatusResponse",
    "Unauthorized",
    "ValidationError",
    "jsonize",
    "LOGGER",
    "INTERNAL_FIELD_VALUE",
    "stdmsgs",
]


def includeme(config: Configurator):
    config.include(".bwcompat")
    config.include(".pyramid")



# pylint: disable=W0622
"""cubicweb_web application packaging information"""


modname = "web"
distname = "cubicweb-web"

numversion = (1, 2, 3)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "Legacy component for web application"
web = "https://forge.extranet.logilab.fr/cubicweb/cubes/web"

__depends__ = {"cubicweb": ">= 4.1.0,<5.0.0", "yams": ">= 0.48"}
__recommends__ = {}

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python :: 3",
    "Programming Language :: JavaScript",
]



# FIXME: VID_BY_MIMETYPE is unfortunately a bit too naive since
#        some browsers (e.g. FF2) send a bunch of mimetypes in
#        the Accept header, for instance:
#          text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,
#          text/plain;q=0.8,image/png,*/*;q=0.5
VID_BY_MIMETYPE = {
    # 'text/xml': 'xml',
    # XXX rss, owl...
}


# this constant is exported to an external module to avoid bugs where the
# backward compatibility in CubicWeb does a lot of import and we end up with
# multiple instances of VID_BY_MIMETYPE while we only want one



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for cubicweb.web.views.entities module"""
from cubicweb import Binary
from cubicweb.devtools import BASE_URL
from cubicweb.entity import can_use_rest_path
from cubicweb.uilib import soup2xhtml
from cubicweb.entities.test.unittest_base import BaseEntityTC

from cubicweb_web.devtools.testlib import WebCWTC


class HTMLtransformTC(BaseEntityTC):
    def test_sanitized_html(self):
        with self.admin_access.repo_cnx() as cnx:
            c = cnx.create_entity(
                "Company",
                name="Babar",
                description="""
Title
=====

Elephant management best practices.

.. raw:: html

   <script>alert("coucou")</script>
""",
                description_format="text/rest",
            )
            cnx.commit()
            c.cw_clear_all_caches()
            self.assertIn(
                "alert", c.printable_value("description", format="text/plain")
            )
            self.assertNotIn(
                "alert", c.printable_value("description", format="text/html")
            )


class EntityCWTC(WebCWTC):
    def setUp(self):
        super().setUp()
        self.backup_dict = {}
        for cls in self.vreg["etypes"].iter_classes():
            self.backup_dict[cls] = (cls.fetch_attrs, cls.cw_fetch_order)

    def tearDown(self):
        super().tearDown()
        for cls in self.vreg["etypes"].iter_classes():
            cls.fetch_attrs, cls.cw_fetch_order = self.backup_dict[cls]

    def test_markdown_printable_value_string(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity(
                "Card",
                title="rest markdown",
                content='This is [an example](http://example.com/ "Title") inline link`',
                content_format="text/markdown",
            )
            self.assertEqual(
                '<p>This is <a href="http://example.com/" '
                'title="Title">an example</a> inline link`</p>',
                e.printable_value("content"),
            )

    def test_printable_value_string(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity(
                "Card",
                title="rest test",
                content="du :eid:`1:*ReST*`",
                content_format="text/rest",
            )
            self.assertEqual(
                e.printable_value("content"),
                f'<p>du <a class="reference" href="{BASE_URL}cwsource/system">*ReST*</a></p>',
            )
            e.cw_attr_cache[
                "content"
            ] = 'du <em>html</em> <ref rql="CWUser X">users</ref>'
            e.cw_attr_cache["content_format"] = "text/html"
            self.assertEqual(
                e.printable_value("content"),
                f'du <em>html</em> <a href="{BASE_URL}view?rql=CWUser%20X">users</a>',
            )
            e.cw_attr_cache["content"] = "du *texte*"
            e.cw_attr_cache["content_format"] = "text/plain"
            self.assertEqual(
                e.printable_value("content").replace("\n", ""), "<p>du *texte*<br/></p>"
            )
            e.cw_attr_cache["title"] = "zou"
            e.cw_attr_cache[
                "content"
            ] = """\
a title
=======
du :eid:`1:*ReST*`"""
            e.cw_attr_cache["content_format"] = "text/rest"
            self.assertEqual(
                e.printable_value("content", format="text/plain"),
                e.cw_attr_cache["content"],
            )

            e.cw_attr_cache["content"] = "<b>yo (zou éà ;)</b>"
            e.cw_attr_cache["content_format"] = "text/html"
            self.assertEqual(
                e.printable_value("content", format="text/plain").strip(),
                "**yo (zou éà ;)**",
            )

    def test_printable_value_bytes(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity(
                "FakeFile",
                data=Binary(b"lambda x: 1"),
                data_format="text/x-python",
                data_encoding="ascii",
                data_name="toto.py",
            )
            from cubicweb import mttransforms

            if mttransforms.HAS_PYGMENTS_TRANSFORMS:
                import pygments

                if tuple(int(i) for i in pygments.__version__.split(".")[:3]) >= (
                    2,
                    1,
                    1,
                ):
                    span = "<span/>"
                else:
                    span = ""
                if tuple(int(i) for i in pygments.__version__.split(".")[:2]) >= (1, 3):
                    mi = "mi"
                else:
                    mi = "mf"

                self.assertEqual(
                    e.printable_value("data"),
                    """<div class="highlight"><pre>%s<span class="k">lambda</span> <span class="n">x</span><span class="p">:</span> <span class="%s">1</span>
</pre></div>"""
                    % (span, mi),
                )
            else:
                self.assertEqual(
                    e.printable_value("data"),
                    """<pre class="python">
    <span style="color: #C00000;">lambda</span> <span style="color: #000000;">x</span><span style="color: #0000C0;">:</span> <span style="color: #0080C0;">1</span>
</pre>""",
                )

            e = req.create_entity(
                "FakeFile",
                data=Binary("*héhéhé*".encode()),
                data_format="text/rest",
                data_encoding="utf-8",
                data_name="toto.txt",
            )
            e = req.entity_from_eid(e.eid)
            self.assertEqual(e.printable_value("data"), "<p><em>héhéhé</em></p>")

        with self.admin_access.web_request() as req:
            e = req.create_entity(
                "FakeFile",
                data=Binary("*héhéhé*".encode()),
                data_format="text/rest",
                data_encoding="utf-8",
                data_name="toto.txt",
            )
            self.assertEqual(e.printable_value("data"), "<p><em>héhéhé</em></p>")

    def test_printable_value_bad_html(self):
        """make sure we don't crash if we try to render invalid XHTML strings"""
        with self.admin_access.web_request() as req:
            e = req.create_entity(
                "Card",
                title="bad html",
                content="<div>R&D<br>",
                content_format="text/html",
            )
            tidy = lambda x: x.replace("\n", "")
            self.assertEqual(
                tidy(e.printable_value("content")), "<div>R&amp;D<br/></div>"
            )
            e.cw_attr_cache["content"] = "yo !! R&D <div> pas fermé"
            self.assertEqual(
                tidy(e.printable_value("content")),
                "yo !! R&amp;D <div> pas fermé</div>",
            )
            e.cw_attr_cache["content"] = "R&D"
            self.assertEqual(tidy(e.printable_value("content")), "R&amp;D")
            e.cw_attr_cache["content"] = "R&D;"
            self.assertEqual(tidy(e.printable_value("content")), "R&amp;D;")
            e.cw_attr_cache["content"] = "yo !! R&amp;D <div> pas fermé"
            self.assertEqual(
                tidy(e.printable_value("content")),
                "yo !! R&amp;D <div> pas fermé</div>",
            )
            e.cw_attr_cache["content"] = "été <div> été"
            self.assertEqual(tidy(e.printable_value("content")), "été <div> été</div>")
            e.cw_attr_cache["content"] = "C&apos;est un exemple s&eacute;rieux"
            self.assertEqual(
                tidy(e.printable_value("content")), "C'est un exemple sérieux"
            )
            # make sure valid xhtml is left untouched
            e.cw_attr_cache["content"] = "<div>R&amp;D<br/></div>"
            self.assertEqual(e.printable_value("content"), e.cw_attr_cache["content"])
            e.cw_attr_cache["content"] = "<div>été</div>"
            self.assertEqual(e.printable_value("content"), e.cw_attr_cache["content"])
            e.cw_attr_cache["content"] = "été"
            self.assertEqual(e.printable_value("content"), e.cw_attr_cache["content"])
            e.cw_attr_cache["content"] = "hop\r\nhop\nhip\rmomo"
            self.assertEqual(e.printable_value("content"), "hop\nhop\nhip\nmomo")

    def test_printable_value_bad_html_ms(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity(
                "Card",
                title="bad html",
                content="<div>R&D<br>",
                content_format="text/html",
            )
            e.cw_attr_cache[
                "content"
            ] = '<div x:foo="bar">ms orifice produces weird html</div>'
            # Caution! current implementation of soup2xhtml strips first div element
            content = soup2xhtml(e.printable_value("content"), "utf-8")
            self.assertMultiLineEqual(
                content, "<div>ms orifice produces weird html</div>"
            )

    def test_fulltextindex(self):
        with self.admin_access.web_request() as req:
            e = self.vreg["etypes"].etype_class("FakeFile")(req)
            e.cw_attr_cache["description"] = "du <em>html</em>"
            e.cw_attr_cache["description_format"] = "text/html"
            e.cw_attr_cache["data"] = Binary(b"some <em>data</em>")
            e.cw_attr_cache["data_name"] = "an html file"
            e.cw_attr_cache["data_format"] = "text/html"
            e.cw_attr_cache["data_encoding"] = "ascii"
            e._cw.transaction_data.clear()
            words = e.cw_adapt_to("IFTIndexable").get_words()
            words["C"].sort()
            self.assertEqual(
                {"C": sorted(["an", "html", "file", "du", "html", "some", "data"])},
                words,
            )

    def test_rest_path(self):
        with self.admin_access.web_request() as req:
            note = req.create_entity("Note", type="z")
            self.assertEqual(note.rest_path(), "note/%s" % note.eid)
            # unique attr
            tag = req.create_entity("Tag", name="x")
            self.assertEqual(tag.rest_path(), "tag/x")
            # test explicit rest_attr
            person = req.create_entity("Personne", prenom="john", nom="doe")
            self.assertEqual(person.rest_path(), "personne/doe")
            # ambiguity test
            person2 = req.create_entity("Personne", prenom="remi", nom="doe")
            person.cw_clear_all_caches()
            self.assertEqual(person.rest_path(), str(person.eid))
            self.assertEqual(person2.rest_path(), str(person2.eid))

            # test explicit int rest_attr
            janze = req.create_entity("Ville", insee=35150)
            self.assertEqual(janze.rest_path(), "ville/35150")

            # unique attr with None value (nom in this case)
            friend = req.create_entity("Ami", prenom="bob")
            self.assertEqual(friend.rest_path(), str(friend.eid))
            # 'ref' below is created without the unique but not required
            # attribute, make sur that the unique _and_ required 'ean' is used
            # as the rest attribute
            ref = req.create_entity("Reference", ean="42-1337-42")
            self.assertEqual(ref.rest_path(), "reference/42-1337-42")

    def test_can_use_rest_path(self):
        self.assertTrue(can_use_rest_path("zobi"))
        # don't use rest if we have /, ? or & in the path (breaks mod_proxy)
        self.assertFalse(can_use_rest_path("zo/bi"))
        self.assertFalse(can_use_rest_path("zo&bi"))
        self.assertFalse(can_use_rest_path("zo?bi"))

    def test_request_cache(self):
        with self.admin_access.web_request() as req:
            user = req.execute('CWUser X WHERE X login "admin"').get_entity(0, 0)
            state = user.in_state[0]
            samestate = req.execute('State X WHERE X name "activated"').get_entity(0, 0)
            self.assertIs(state, samestate)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



import unittest

from cubicweb_web import http_headers


class TestGenerators(unittest.TestCase):
    def test_generate_true_false(self):
        for v in (True, 1, "true", "True", "TRUE"):
            self.assertEqual("true", http_headers.generateTrueFalse(v))
        for v in (False, 0, "false", "False", "FALSE"):
            self.assertEqual("false", http_headers.generateTrueFalse(v))

        with self.assertRaises(ValueError):
            http_headers.generateTrueFalse("any value")


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""base xml and rss views"""

from base64 import b64encode
from time import timezone

from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import (
    is_instance,
    non_final_entity,
    one_line_rset,
    appobject_selectable,
    adaptable,
)
from cubicweb.uilib import simple_sgml_tag
from cubicweb_web.view import EntityView, AnyRsetView, Component
from cubicweb_web.views.navigation import limited_rql
from cubicweb_web import httpcache, component


def encode_bytes(value):
    return "<![CDATA[%s]]>" % b64encode(value.getvalue())


# see cubicweb.sobjects.parser.DEFAULT_CONVERTERS
SERIALIZERS = {
    "String": xml_escape,
    "Bytes": encode_bytes,
    "Date": lambda x: x.strftime("%Y-%m-%d"),
    "Datetime": lambda x: x.strftime("%Y-%m-%d %H:%M:%S"),
    "Time": lambda x: x.strftime("%H:%M:%S"),
    "TZDatetime": lambda x: x.strftime("%Y-%m-%d %H:%M:%S"),  # XXX TZ
    "TZTime": lambda x: x.strftime("%H:%M:%S"),
    "Interval": lambda x: x.days * 60 * 60 * 24 + x.seconds,
}

# base xml views ##############################################################


class XMLView(EntityView):
    """xml view for entities"""

    __regid__ = "xml"
    title = _("xml export (entities)")
    templatable = False
    content_type = "text/xml"
    xml_root = "rset"
    item_vid = "xmlitem"

    def cell_call(self, row, col):
        self.wview(self.item_vid, self.cw_rset, row=row, col=col)

    def call(self):
        """display a list of entities by calling their <item_vid> view"""
        self.w('<?xml version="1.0" encoding="%s"?>\n', self._cw.encoding)
        self.w('<%s size="%s">\n', self.xml_root, len(self.cw_rset))
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self.w("</%s>\n", self.xml_root)


class XMLItemView(EntityView):
    __regid__ = "xmlitem"

    def entity_call(self, entity):
        """element as an item for an xml feed"""
        entity.complete()
        source = entity.cw_source[0].name
        self.w(
            '<%s eid="%s" cwuri="%s" cwsource="%s">\n',
            entity.cw_etype,
            entity.eid,
            entity.cwuri,
            source,
        )
        for rschema, attrschema in sorted(entity.e_schema.attribute_definitions()):
            attr = rschema.type
            if attr in ("eid", "cwuri"):
                continue
            else:
                try:
                    value = entity.cw_attr_cache[attr]
                except KeyError:
                    # Bytes
                    continue
            if value is None:
                self.w("  <%s/>\n", attr)
            else:
                if attrschema in SERIALIZERS:
                    value = SERIALIZERS[attrschema](value)
                self.w("  <%s>%s</%s>\n", attr, value, attr)
        for relstr in self._cw.list_form_param("relation"):
            try:
                rtype, role = relstr.split("-")
            except ValueError:
                self.error("badly formated relation name %r", relstr)
                continue
            if role == "subject":
                getrschema = entity.e_schema.subject_relations
            elif role == "object":
                getrschema = entity.e_schema.object_relations
            else:
                self.error("badly formated relation name %r", relstr)
                continue
            if rtype not in getrschema:
                self.error("unexisting relation %r", relstr)
                continue
            self.w('  <%s role="%s">\n', rtype, role)
            self.wview("xmlrelateditem", entity.related(rtype, role, safe=True), "null")
            self.w("  </%s>\n", rtype)
        self.w("</%s>\n", entity.e_schema)


class XMLRelatedItemView(EntityView):
    __regid__ = "xmlrelateditem"
    add_div_section = False

    def entity_call(self, entity):
        # XXX put unique attributes as xml attribute, they are much probably
        # used to search existing entities in client data feed, and putting it
        # here may avoid an extra request to get those attributes values
        self.w(
            '    <%s eid="%s" cwuri="%s"/>\n', entity.e_schema, entity.eid, entity.cwuri
        )


class XMLRelatedItemStateView(XMLRelatedItemView):
    __select__ = is_instance("State")

    def entity_call(self, entity):
        self.w(
            '    <%s eid="%s" cwuri="%s" name="%s"/>\n',
            entity.e_schema,
            entity.eid,
            entity.cwuri,
            entity.name,
        )


class XMLRsetView(AnyRsetView):
    """dumps raw rset as xml"""

    __regid__ = "rsetxml"
    title = _("xml export")
    templatable = False
    content_type = "text/xml"
    xml_root = "rset"

    def call(self):
        rset, descr = self.cw_rset, self.cw_rset.description
        entity_schema_for = self._cw.vreg.schema.entity_schema_for
        labels = self.columns_labels(tr=False)
        self.w('<?xml version="1.0" encoding="%s"?>\n', self._cw.encoding)
        self.w('<%s query="%s">\n', self.xml_root, rset.printable_rql())
        for rowindex, row in enumerate(self.cw_rset):
            self.w(" <row>\n")
            for colindex, val in enumerate(row):
                etype = descr[rowindex][colindex]
                tag = labels[colindex]
                attrs = {}
                if "(" in tag:
                    attrs["expr"] = tag
                    tag = "funccall"
                if val is not None and not entity_schema_for(etype).final:
                    attrs["eid"] = val
                    # csvrow.append(val) # val is eid in that case
                    val = self._cw.view(
                        "textincontext", rset, row=rowindex, col=colindex
                    )
                else:
                    val = self._cw.view(
                        "final", rset, row=rowindex, col=colindex, format="text/plain"
                    )
                self.w(simple_sgml_tag(tag, val, **attrs))
            self.w("\n </row>\n")
        self.w("</%s>\n", self.xml_root)


# RSS stuff ###################################################################


class IFeedAdapter(EntityAdapter):
    __needs_bw_compat__ = True
    __regid__ = "IFeed"
    __select__ = is_instance("Any")

    def rss_feed_url(self):
        """return a URL to the rss feed for this entity"""
        return self.entity.absolute_url(vid="rss")


class RSSFeedURL(Component):
    __regid__ = "rss_feed_url"
    __select__ = non_final_entity()

    def feed_url(self):
        return self._cw.build_url(rql=limited_rql(self.cw_rset), vid="rss")


class RSSEntityFeedURL(Component):
    __regid__ = "rss_feed_url"
    __select__ = one_line_rset() & adaptable("IFeed")

    def feed_url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.cw_adapt_to("IFeed").rss_feed_url()


class RSSIconBox(component.CtxComponent):
    """just display the RSS icon on uniform result set"""

    __regid__ = "rss"
    __select__ = component.CtxComponent.__select__ & appobject_selectable(
        "components", "rss_feed_url"
    )

    visible = False
    order = 999

    def render(self, w, **kwargs):
        try:
            rss = self._cw.uiprops["RSS_LOGO"]
        except KeyError:
            self.error("missing RSS_LOGO external resource")
            return
        urlgetter = self._cw.vreg["components"].select(
            "rss_feed_url", self._cw, rset=self.cw_rset
        )
        url = urlgetter.feed_url()
        w('<a href="%s"><img src="%s" alt="rss"/></a>\n', url, rss)


class RSSView(XMLView):
    __regid__ = "rss"
    title = _("rss export")
    templatable = False
    content_type = "text/xml"
    http_cache_manager = httpcache.MaxAgeHTTPCacheManager
    cache_max_age = 60 * 60 * 2  # stay in http cache for 2 hours by default
    item_vid = "rssitem"

    def _open(self):
        req = self._cw
        self.w('<?xml version="1.0" encoding="%s"?>\n', req.encoding)
        self.w('<rss version="2.0" xmlns:dc="http://purl.org/dc/elements/1.1/">\n')
        self.w("  <channel>\n")
        self.w("    <title>%s RSS Feed</title>\n", self.page_title())
        self.w("    <description>%s</description>\n", req.form.get("vtitle", ""))
        params = req.form.copy()
        params.pop("vid", None)
        self.w("    <link>%s</link>\n", self._cw.build_url(**params))

    def _close(self):
        self.w("  </channel>\n")
        self.w("</rss>")

    def call(self):
        """display a list of entities by calling their <item_vid> view"""
        self._open()
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self._close()

    def cell_call(self, row, col):
        self.wview(self.item_vid, self.cw_rset, row=row, col=col)


class RSSItemView(EntityView):
    __regid__ = "rssitem"
    date_format = "%%Y-%%m-%%dT%%H:%%M%+03i:00" % (timezone / 3600)
    add_div_section = False

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        self.w("<item>\n")
        self.w('<guid isPermaLink="true">%s</guid>\n', entity.absolute_url())
        self.render_title_link(entity)
        self.render_description(entity)
        self._marker("dc:date", entity.dc_date(self.date_format))
        self.render_entity_creator(entity)
        self.w("</item>\n")

    def render_description(self, entity):
        self._marker("description", entity.dc_description(format="text/html"))

    def render_title_link(self, entity):
        self._marker("title", entity.dc_long_title())
        self._marker("link", entity.absolute_url())

    def render_entity_creator(self, entity):
        if entity.creator:
            self._marker("dc:creator", entity.dc_creator())

    def _marker(self, marker, value):
        if value:
            self.w("  <%s>%s</%s>\n", marker, value, marker)



import json

from cubicweb.utils import UStringIO
from logilab.mtconverter import xml_escape


class HTMLHead(UStringIO):
    """wraps HTML header's stream

    Request objects use a HTMLHead instance to ease adding of
    javascripts and stylesheets
    """

    js_unload_code = """if (typeof(pageDataUnloaded) == 'undefined') {
    jQuery(window).unload(unloadPageData);
    pageDataUnloaded = true;
}"""
    script_opening = '<script type="text/javascript">\n'
    script_closing = "\n</script>"

    def __init__(self, req, *args, **kwargs):
        super(HTMLHead, self).__init__(*args, **kwargs)
        self.jsvars = []
        self.jsfiles = []
        self.cssfiles = []
        self.post_inlined_scripts = []
        self.pagedata_unload = False
        self._cw = req
        self.datadir_url = req.datadir_url

    def add_raw(self, rawheader):
        self.write(rawheader)

    def define_var(self, var, value, override=True):
        """adds a javascript var declaration / assginment in the header

        :param var: the variable name
        :param value: the variable value (as a raw python value,
                      it will be jsonized later)
        :param override: if False, don't set the variable value if the variable
                         is already defined. Default is True.
        """
        self.jsvars.append((var, value, override))

    def add_post_inline_script(self, content):
        self.post_inlined_scripts.append(content)

    def add_onload(self, jscode):
        self.add_post_inline_script(
            """$(cw).one('server-response', function(event) {
%s});"""
            % jscode
        )

    def add_js(self, jsfile, script_attributes={}):
        """adds `jsfile` to the list of javascripts used in the webpage

        This function checks if the file has already been added
        :param jsfile: the script's URL
        :param script_attributes: a dictionary of <script> attributes
            e.g. {"defer": True, "integrity": "sha256-..."}
            The attributes are not taken into account if the global option
            'concat-resources' is set to True
        """
        # XXX: if a script is added twice with different attributes, only the first
        #   call to add this script will be taken into account
        if jsfile not in [_jsfile["src"] for _jsfile in self.jsfiles]:
            self.jsfiles.append({"src": jsfile, **script_attributes})

    def add_css(self, cssfile, media="all"):
        """adds `cssfile` to the list of javascripts used in the webpage

        This function checks if the file has already been added
        :param cssfile: the stylesheet's URL
        """
        if (cssfile, media) not in self.cssfiles:
            self.cssfiles.append((cssfile, media))

    def add_unload_pagedata(self):
        """registers onunload callback to clean page data on server"""
        if not self.pagedata_unload:
            self.post_inlined_scripts.append(self.js_unload_code)
            self.pagedata_unload = True

    def concat_urls(self, urls):
        """concatenates urls into one url usable by Apache mod_concat

        This method returns the url without modifying it if there is only
        one element in the list
        :param urls: list of local urls/filenames to concatenate
        """
        if len(urls) == 1:
            return urls[0]
        len_prefix = len(self.datadir_url)
        concated = ",".join(url[len_prefix:] for url in urls)
        return "%s??%s" % (self.datadir_url, concated)

    def group_urls(self, urls_spec):
        """parses urls_spec in order to generate concatenated urls
        for js and css includes

        This method checks if the file is local and if it shares options
        with direct neighbors
        :param urls_spec: entire list of urls/filenames to inspect
        """
        concatable = []
        prev_islocal = False
        prev_key = None
        for url, key in urls_spec:
            islocal = url.startswith(self.datadir_url)
            if concatable and (islocal != prev_islocal or key != prev_key):
                yield (self.concat_urls(concatable), prev_key)
                del concatable[:]
            if not islocal:
                yield (url, key)
            else:
                concatable.append(url)
            prev_islocal = islocal
            prev_key = key
        if concatable:
            yield (self.concat_urls(concatable), prev_key)

    def getvalue(self, skiphead=False):
        """reimplement getvalue to provide a consistent (and somewhat browser
        optimzed cf. http://stevesouders.com/cuzillion) order in external
        resources declaration
        """
        w = self.write
        # 1/ variable declaration if any
        if self.jsvars:
            if skiphead:
                w("<cubicweb:script>")
            else:
                w(self.script_opening)
            for var, value, override in self.jsvars:
                vardecl = "%s = %s;" % (var, json.dumps(value))
                if not override:
                    vardecl = 'if (typeof %s == "undefined") {%s}' % (var, vardecl)
                w(vardecl + "\n")
            if skiphead:
                w("</cubicweb:script>")
            else:
                w(self.script_closing)
        # 2/ css files
        if self.datadir_url and self._cw.vreg.config["concat-resources"]:
            cssfiles = self.group_urls(self.cssfiles)
            jsfiles = (
                {"src": src}
                for src, _ in self.group_urls(
                    (jsfile["src"], None) for jsfile in self.jsfiles
                )
            )
        else:
            cssfiles = self.cssfiles
            jsfiles = self.jsfiles
        for cssfile, media in cssfiles:
            w(
                '<link rel="stylesheet" type="text/css" media="%s" href="%s"/>\n'
                % (media, xml_escape(cssfile))
            )
        # 3/ js files
        for jsfile in jsfiles:
            script_attributes = [
                f'{key}="{val}"' if val is not True else f"{key}"
                for key, val in jsfile.items()
            ]
            if skiphead:
                # Don't insert <script> tags directly as they would be
                # interpreted directly by some browsers (e.g. IE).
                # Use <cubicweb:script> tags instead and let
                # `loadAjaxHtmlHead` handle the script insertion / execution.
                w(
                    '<cubicweb:script src="%s"></cubicweb:script>\n'
                    % xml_escape(jsfile["src"])
                )
                # FIXME: a probably better implementation might be to add
                #        JS or CSS urls in a JS list that loadAjaxHtmlHead
                #        would iterate on and postprocess:
                #            cw._ajax_js_scripts.push('myscript.js')
                #        Then, in loadAjaxHtmlHead, do something like:
                #            jQuery.each(cw._ajax_js_script, jQuery.getScript)
            else:
                w(
                    '<script type="text/javascript" %s></script>\n'
                    % " ".join(script_attributes)
                )
        # 4/ post inlined scripts (i.e. scripts depending on other JS files)
        if self.post_inlined_scripts:
            if skiphead:
                for script in self.post_inlined_scripts:
                    w("<cubicweb:script>")
                    w(xml_escape(script))
                    w("</cubicweb:script>")
            else:
                w(self.script_opening)
                w("\n\n".join(self.post_inlined_scripts))
                w(self.script_closing)
        # at the start of this function, the parent UStringIO may already have
        # data in it, so we can't w(u'<head>\n') at the top. Instead, we create
        # a temporary UStringIO to get the same debugging output formatting
        # if debugging is enabled.
        headtag = UStringIO(tracewrites=self.tracewrites)
        if not skiphead:
            headtag.write("<head>\n")
            w("</head>\n")
        return headtag.getvalue() + super(HTMLHead, self).getvalue()


class HTMLStream:
    """represents a HTML page.

    This is used my main templates so that HTML headers can be added
    at any time during the page generation.

    HTMLStream uses the (U)StringIO interface to be compliant with
    existing code.
    """

    def __init__(self, req):
        self.tracehtml = req.tracehtml
        # stream for <head>
        self.head = req.html_headers
        # main stream
        self.body = UStringIO(tracewrites=req.tracehtml)
        # this method will be assigned to self.w in views
        self.write = self.body.write
        self.doctype = ""
        self._htmlattrs = [("lang", req.lang)]
        # keep main_stream's reference on req for easier text/html demoting
        req.main_stream = self

    def add_htmlattr(self, attrname, attrvalue):
        self._htmlattrs.append((attrname, attrvalue))

    def set_htmlattrs(self, attrs):
        self._htmlattrs = attrs

    def set_doctype(self, doctype):
        self.doctype = doctype

    @property
    def htmltag(self):
        attrs = " ".join(
            '%s="%s"' % (attr, xml_escape(value)) for attr, value in self._htmlattrs
        )
        if attrs:
            return '<html xmlns:cubicweb="http://www.cubicweb.org" %s>' % attrs
        return '<html xmlns:cubicweb="http://www.cubicweb.org">'

    def getvalue(self):
        """writes HTML headers, closes </head> tag and writes HTML body"""
        if self.tracehtml:
            css = "\n".join(
                (
                    "span {",
                    "  font-family: monospace;",
                    "  word-break: break-all;",
                    "  word-wrap: break-word;",
                    "}",
                    "span:hover {",
                    "  color: red;",
                    "  text-decoration: underline;",
                    "}",
                )
            )
            style = '<style type="text/css">\n%s\n</style>\n' % css
            return (
                "<!DOCTYPE html>\n"
                + "<html>\n<head>\n%s\n</head>\n" % style
                + "<body>\n"
                + "<span>"
                + xml_escape(self.doctype)
                + "</span><br/>"
                + "<span>"
                + xml_escape(self.htmltag)
                + "</span><br/>"
                + self.head.getvalue()
                + self.body.getvalue()
                + "<span>"
                + xml_escape("</html>")
                + "</span>"
                + "</body>\n</html>"
            )
        return "%s\n%s\n%s\n%s\n</html>" % (
            self.doctype,
            self.htmltag,
            self.head.getvalue(),
            self.body.getvalue(),
        )


def url_path_starts_with_prefix(url_path, prefix):
    return url_path.lstrip("/").startswith(prefix.lstrip("/"))


def remove_prefix_from_url_path(url_path, prefix):
    return url_path.lstrip("/")[len(prefix.lstrip("/")) :]



# copyright 2012-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
from functools import partial

from pytz import utc

from cubicweb.devtools.testlib import real_error_handling
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import is_instance
from cubicweb_web import http_headers
from cubicweb_web.devtools.testlib import WebCWTC


class IDownloadableUser(EntityAdapter):
    __regid__ = "IDownloadable"
    __select__ = is_instance("CWUser")

    def download_content_type(self):
        """return MIME type of the downloadable content"""
        return "text/plain"

    def download_encoding(self):
        """return encoding of the downloadable content"""
        return "ascii"

    def download_file_name(self):
        """return file name of the downloadable content"""
        return self.entity.name() + ".txt"

    def download_data(self):
        return b"Babar is not dead!"


class BrokenIDownloadableGroup(IDownloadableUser):
    __regid__ = "IDownloadable"
    __select__ = is_instance("CWGroup")

    def download_file_name(self):
        return self.entity.name + ".txt"

    def download_data(self):
        raise OSError()


class IDownloadableCWTC(WebCWTC):
    def setUp(self):
        super().setUp()
        self.vreg.register(IDownloadableUser)
        self.addCleanup(partial(self.vreg.unregister, IDownloadableUser))

    def test_header_simple_case(self):
        with self.admin_access.web_request() as req:
            req.form["vid"] = "download"
            req.form["eid"] = str(req.user.eid)
            data = self.ctrl_publish(req, "view")
            get = req.headers_out.getRawHeaders
            self.assertEqual(
                ['attachment;filename="admin.txt"'], get("content-disposition")
            )
            self.assertEqual(["text/plain;charset=ascii"], get("content-type"))
            last_mod = http_headers.parseDateTime(get("last-modified")[0])
            self.assertEqual(
                datetime.fromtimestamp(last_mod, tz=utc),
                req.user.modification_date.replace(microsecond=0),
            )
            self.assertEqual(b"Babar is not dead!", data)

    def test_header_with_space(self):
        with self.admin_access.web_request() as req:
            self.create_user(req, login="c c l a", password="babar")
            req.cnx.commit()
        with self.new_access("c c l a").web_request() as req:
            req.form["vid"] = "download"
            req.form["eid"] = str(req.user.eid)
            data = self.ctrl_publish(req, "view")
            get = req.headers_out.getRawHeaders
            self.assertEqual(
                ['attachment;filename="c c l a.txt"'], get("content-disposition")
            )
            self.assertEqual(["text/plain;charset=ascii"], get("content-type"))
            self.assertEqual(b"Babar is not dead!", data)

    def test_header_with_space_and_comma(self):
        with self.admin_access.web_request() as req:
            self.create_user(req, login='c " l\\ a', password="babar")
            req.cnx.commit()
        with self.new_access('c " l\\ a').web_request() as req:
            req.form["vid"] = "download"
            req.form["eid"] = str(req.user.eid)
            data = self.ctrl_publish(req, "view")
            get = req.headers_out.getRawHeaders
            self.assertEqual(
                [r'attachment;filename="c \" l\\ a.txt"'], get("content-disposition")
            )
            self.assertEqual(["text/plain;charset=ascii"], get("content-type"))
            self.assertEqual(b"Babar is not dead!", data)

    def test_header_unicode_filename(self):
        with self.admin_access.web_request() as req:
            self.create_user(req, login="cécilia", password="babar")
            req.cnx.commit()
        with self.new_access("cécilia").web_request() as req:
            req.form["vid"] = "download"
            req.form["eid"] = str(req.user.eid)
            self.ctrl_publish(req, "view")
            get = req.headers_out.getRawHeaders
            self.assertEqual(
                [
                    """attachment;filename="ccilia.txt";filename*=utf-8''c%C3%A9cilia.txt"""
                ],
                get("content-disposition"),
            )

    def test_header_unicode_long_filename(self):
        name = "Bèrte_hô_grand_nôm_ça_va_totallement_déborder_de_la_limite_là"
        with self.admin_access.web_request() as req:
            self.create_user(req, login=name, password="babar")
            req.cnx.commit()
        with self.new_access(name).web_request() as req:
            req.form["vid"] = "download"
            req.form["eid"] = str(req.user.eid)
            self.ctrl_publish(req, "view")
            get = req.headers_out.getRawHeaders
            self.assertEqual(
                [
                    """attachment;filename="Brte_h_grand_nm_a_va_totallement_dborder_de_la_limite_l.txt";filename*=utf-8''B%C3%A8rte_h%C3%B4_grand_n%C3%B4m_%C3%A7a_va_totallement_d%C3%A9border_de_la_limite_l%C3%A0.txt"""
                ],
                get("content-disposition"),
            )

    def test_download_data_error(self):
        self.vreg.register(BrokenIDownloadableGroup)
        self.addCleanup(partial(self.vreg.unregister, BrokenIDownloadableGroup))
        with self.admin_access.web_request() as req:
            req.form["vid"] = "download"
            req.form["eid"] = str(
                req.execute('CWGroup X WHERE X name "managers"')[0][0]
            )
            with real_error_handling(self.app):
                self.app_handle_request(req)
            get = req.headers_out.getRawHeaders
            self.assertEqual(["text/html;charset=UTF-8"], get("content-type"))
            self.assertEqual(None, get("content-disposition"))
            self.assertEqual(req.status_out, 500)


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2010-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittests for cw.web.formwidgets"""

from logilab.common.testlib import unittest_main, mock_object as mock

from cubicweb_web import formwidgets, formfields
from cubicweb_web.devtools.testlib import WebCWTC, FakeRequest
from cubicweb_web.views.forms import FieldsForm


class WidgetsCWTC(WebCWTC):
    def test_editableurl_widget(self):
        field = formfields.guess_field(self.schema["Bookmark"], self.schema["path"])
        widget = formwidgets.EditableURLWidget()
        req = FakeRequest(form={"path-subjectfqs:A": "param=value&vid=view"})
        form = mock(_cw=req, formvalues={}, edited_entity=mock(eid="A"))
        self.assertEqual(
            widget.process_field_data(form, field), "?param=value%26vid%3Dview"
        )

    def test_bitselect_widget(self):
        field = formfields.guess_field(
            self.schema["CWAttribute"], self.schema["ordernum"]
        )
        field.choices = [
            (
                "un",
                "1",
            ),
            (
                "deux",
                "2",
            ),
        ]
        widget = formwidgets.BitSelect()
        req = FakeRequest(form={"ordernum-subject:A": ["1", "2"]})
        form = mock(
            _cw=req, formvalues={}, edited_entity=mock(eid="A"), form_previous_values=()
        )
        self.assertMultiLineEqual(
            widget._render(form, field, None),
            """\
<select id="ordernum-subject:A" multiple="multiple" name="ordernum-subject:A" size="2">
<option selected="selected" value="2">deux</option>
<option selected="selected" value="1">un</option>
</select>""",
        )
        self.assertEqual(widget.process_field_data(form, field), 3)

    def test_xml_escape_checkbox(self):
        class TestForm(FieldsForm):
            bool = formfields.BooleanField(
                ignore_req_params=True,
                choices=[("python >> others", "1")],
                widget=formwidgets.CheckBox(),
            )

        with self.admin_access.web_request() as req:
            form = TestForm(req, None)
            form.build_context()
            field = form.field_by_name("bool")
            widget = field.widget
            self.assertMultiLineEqual(
                widget._render(form, field, None),
                '<label><input id="bool" name="bool" '
                'type="checkbox" value="1" />&#160;'
                "python &gt;&gt; others</label>",
            )


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittests for cw.web.formfields"""

from logilab.common.testlib import TestCase, unittest_main, mock_object as mock

import cubicweb

from cubicweb_web.devtools.testlib import WebCWTC, WebApptestConfiguration
from cubicweb_web.formfields import (
    guess_field,
    StringField,
    RichTextField,
    DateTimeField,
    RelationField,
    FileField,
    SizeConstraint,
    StaticVocabularyConstraint,
    BooleanField,
    IntField,
    CompoundField,
    vocab_sort,
)
from cubicweb_web.formwidgets import PasswordInput, Select, Radio
from cubicweb_web.views.forms import EntityFieldsForm, FieldsForm


def setUpModule(*args):
    global schema
    config = WebApptestConfiguration("data", __file__)
    config.bootstrap_cubes()
    schema = config.load_schema()


class GuessFieldCWTC(WebCWTC):
    def test_state_fields(self):
        with self.admin_access.web_request() as req:
            title_field = guess_field(schema["State"], schema["name"], req=req)
            self.assertIsInstance(title_field, StringField)
            self.assertEqual(title_field.required, True)

        with self.admin_access.web_request() as req:
            description_field = guess_field(
                schema["State"], schema["description"], req=req
            )
            self.assertIsInstance(description_field, RichTextField)
            self.assertEqual(description_field.required, False)
            self.assertEqual(description_field.format_field, None)

        # description_format_field = guess_field(schema['State'], schema['description_format'])
        # self.assertEqual(description_format_field, None)

        with self.admin_access.web_request() as req:
            description_format_field = guess_field(
                schema["State"], schema["description_format"], req=req
            )
            self.assertEqual(description_format_field.internationalizable, True)
            self.assertEqual(description_format_field.sort, True)

    def test_cwuser_fields(self):
        with self.admin_access.web_request() as req:
            upassword_field = guess_field(
                schema["CWUser"], schema["upassword"], req=req
            )
            self.assertIsInstance(upassword_field, StringField)
            self.assertIsInstance(upassword_field.widget, PasswordInput)
            self.assertEqual(upassword_field.required, True)

        with self.admin_access.web_request() as req:
            last_login_time_field = guess_field(
                schema["CWUser"], schema["last_login_time"], req=req
            )
            self.assertIsInstance(last_login_time_field, DateTimeField)
            self.assertEqual(last_login_time_field.required, False)

        with self.admin_access.web_request() as req:
            in_group_field = guess_field(schema["CWUser"], schema["in_group"], req=req)
            self.assertIsInstance(in_group_field, RelationField)
            self.assertEqual(in_group_field.required, True)
            self.assertEqual(in_group_field.role, "subject")
            self.assertEqual(
                in_group_field.help, "groups grant permissions to the user"
            )

        with self.admin_access.web_request() as req:
            owned_by_field = guess_field(
                schema["CWUser"], schema["owned_by"], "object", req=req
            )
            self.assertIsInstance(owned_by_field, RelationField)
            self.assertEqual(owned_by_field.required, False)
            self.assertEqual(owned_by_field.role, "object")

    def test_file_fields(self):
        with self.admin_access.web_request() as req:
            data_field = guess_field(schema["File"], schema["data"], req=req)
            self.assertIsInstance(data_field, FileField)
            self.assertEqual(data_field.required, True)
            self.assertIsInstance(data_field.format_field, StringField)
            self.assertIsInstance(data_field.encoding_field, StringField)
            self.assertIsInstance(data_field.name_field, StringField)

    def test_constraints_priority(self):
        with self.admin_access.web_request() as req:
            salesterm_field = guess_field(
                schema["Salesterm"], schema["reason"], req=req
            )
            constraints = (
                schema["reason"].relation_definition("Salesterm", "String").constraints
            )
            self.assertEqual(
                [c.__class__ for c in constraints],
                [SizeConstraint, StaticVocabularyConstraint],
            )
            self.assertIsInstance(salesterm_field, StringField)
            self.assertIsInstance(salesterm_field.widget, Select)

    def test_bool_field_base(self):
        with self.admin_access.web_request() as req:
            field = guess_field(schema["CWAttribute"], schema["indexed"], req=req)
            self.assertIsInstance(field, BooleanField)
            self.assertEqual(field.required, False)
            self.assertIsInstance(field.widget, Radio)
            self.assertEqual(
                field.vocabulary(mock(_cw=mock(_=cubicweb._))),
                [("yes", "1"), ("no", "")],
            )

    def test_bool_field_explicit_choices(self):
        with self.admin_access.web_request() as req:
            field = guess_field(
                schema["CWAttribute"],
                schema["indexed"],
                choices=[("maybe", "1"), ("no", "")],
                req=req,
            )
            self.assertIsInstance(field.widget, Radio)
            self.assertEqual(
                field.vocabulary(mock(req=mock(_=cubicweb._))),
                [("maybe", "1"), ("no", "")],
            )


class MoreFieldsCWTC(WebCWTC):
    def test_rtf_format_field(self):
        with self.admin_access.web_request() as req:
            req.use_fckeditor = lambda: False
            e = self.vreg["etypes"].etype_class("State")(req)
            form = EntityFieldsForm(req, entity=e)
            description_field = guess_field(schema["State"], schema["description"])
            description_format_field = description_field.get_format_field(form)
            self.assertEqual(description_format_field.internationalizable, True)
            self.assertEqual(description_format_field.sort, True)
            # unlike below, initial is bound to form.form_field_format
            self.assertEqual(description_format_field.value(form), "text/plain")
            req.cnx.create_entity(
                "CWProperty",
                pkey="ui.default-text-format",
                value="text/rest",
                for_user=req.user.eid,
            )
            req.cnx.commit()
        with self.admin_access.web_request() as req:
            form = EntityFieldsForm(req, entity=e)
            self.assertEqual(description_format_field.value(form), "text/rest")

    def test_property_key_field(self):
        from cubicweb_web.views.cwproperties import PropertyKeyField

        with self.admin_access.web_request() as req:
            field = PropertyKeyField(name="test")
            e = self.vreg["etypes"].etype_class("CWProperty")(req)
            renderer = self.vreg["formrenderers"].select("base", req)
            form = EntityFieldsForm(req, entity=e)
            form.formvalues = {}
            field.render(form, renderer)


class CompoundFieldCWTC(WebCWTC):
    def test_multipart(self):
        """Ensures that compound forms have needs_multipart set if their children require it"""

        class AForm(FieldsForm):
            comp = CompoundField([IntField(), StringField()])

        with self.admin_access.web_request() as req:
            aform = AForm(req, None)
            self.assertFalse(aform.needs_multipart)

        class MForm(FieldsForm):
            comp = CompoundField([IntField(), FileField()])

        with self.admin_access.web_request() as req:
            mform = MForm(req, None)
            self.assertTrue(mform.needs_multipart)


class UtilsTC(TestCase):
    def test_vocab_sort(self):
        self.assertEqual(
            vocab_sort(
                [
                    ("Z", 1),
                    ("A", 2),
                    ("Group 1", None),
                    ("Y", 3),
                    ("B", 4),
                    ("Group 2", None),
                    ("X", 5),
                    ("C", 6),
                ]
            ),
            [
                ("A", 2),
                ("Z", 1),
                ("Group 1", None),
                ("B", 4),
                ("Y", 3),
                ("Group 2", None),
                ("C", 6),
                ("X", 5),
            ],
        )


if __name__ == "__main__":
    unittest_main()



from cubicweb_web.devtools.testlib import PyramidWebCWTC


def redirect_rule(req, url, matchdict):
    return f"{url}/test"


class UrlRedirectTC(PyramidWebCWTC):
    def includeme(self, config):
        config.add_redirection_rule(r"^/\w+$", redirect_rule)
        config.add_redirection_rule(
            r"^/do-not-keep-query", redirect_rule, keep_query_component=False
        )

    def test_redirect_rules(self):
        res = self.webapp.get("/123456", status="*")
        self.assertEqual(res.status_int, 303)
        self.assertEqual(res.location, f"{res.request.application_url}/123456/test")

    def test_redirect_rules_with_param(self):
        res = self.webapp.get(
            "/123456", params={"poulet": "frie", "frite": "non"}, status="*"
        )
        self.assertEqual(res.status_int, 303)
        self.assertEqual(
            res.location,
            f"{res.request.application_url}/123456/test?poulet=frie&frite=non",
        )
        res = self.webapp.get(
            "/do-not-keep-query", params={"poulet": "frie", "frite": "non"}, status="*"
        )
        self.assertEqual(res.status_int, 303)
        self.assertEqual(
            res.location,
            f"{res.request.application_url}/do-not-keep-query/test",
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams.buildobjs import (
    EntityType,
    RelationDefinition,
    ComputedRelation,
    SubjectRelation,
    String,
    Int,
    Datetime,
    Boolean,
    Float,
    TZDatetime,
    Bytes,
    RichString,
)
from yams.constraints import IntervalBoundConstraint

from cubicweb import _
from cubicweb.schema import (
    RRQLExpression,
    RQLConstraint,
    RQLVocabularyConstraint,
    WorkflowableEntityType,
)


def rperms(var, read=("managers", "users")):
    return {
        "read": read,
        "add": ("managers", RRQLExpression(f"U has_update_permission {var}")),
        "delete": (
            "managers",
            RRQLExpression(f"U has_delete_permission {var}"),
        ),
    }


class Salesterm(EntityType):
    described_by_test = SubjectRelation(
        "File", cardinality="1*", composite="subject", inlined=True
    )
    amount = Int(constraints=[IntervalBoundConstraint(0, 100)])
    reason = String(maxsize=20, vocabulary=["canceled", "sold"])


class tags(RelationDefinition):
    subject = "Tag"
    object = ("BlogEntry", "CWUser", "Ami", "Personne", "Note")


class checked_by(RelationDefinition):
    subject = "BlogEntry"
    object = "CWUser"
    cardinality = "?*"
    __permissions__ = {
        "add": ("managers",),
        "read": ("managers", "users"),
        "delete": ("managers",),
    }


class Personne(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (
            "managers",
            "validators",
            "users",
        ),
        "update": (
            "managers",
            "validators",
            "owners",
        ),
        "delete": (
            "managers",
            "validators",
            "owners",
        ),
    }
    nom = String(fulltextindexed=True, required=True)
    prenom = String(fulltextindexed=True)
    type = String()
    sexe = String(
        maxsize=1,
        default="M",
        __permissions__={
            "read": (
                "managers",
                "users",
                "guests",
            ),
            "add": ("managers", "users"),
            "update": ("managers",),
        },
    )
    promo = String(vocabulary=("bon", "pasbon"))
    titre = String(fulltextindexed=True, maxsize=128)
    ass = String(maxsize=128)
    web = String(maxsize=128)
    tel = Int()
    fax = Int()
    datenaiss = Datetime()
    tzdatenaiss = TZDatetime()
    test = Boolean()
    description = String()
    salary = Float()
    travaille = SubjectRelation("Societe")
    tags = SubjectRelation("BlogEntry")
    evaluee = SubjectRelation(("Note", "Personne"))
    connait = SubjectRelation(
        "Personne",
        symmetric=True,
        constraints=[
            RQLConstraint("NOT S identity O"),
            # conflicting constraints, see cw_unrelated_rql tests in
            # unittest_entity.py
            RQLVocabularyConstraint('NOT (S connait P, P nom "toto")'),
            RQLVocabularyConstraint('S travaille P, P nom "tutu"'),
        ],
    )
    actionnaire = SubjectRelation(
        "Societe",
        cardinality="??",
        constraints=[RQLConstraint("NOT EXISTS(O contrat_exclusif S)")],
    )
    dirige = SubjectRelation(
        "Societe", cardinality="??", constraints=[RQLConstraint("S actionnaire O")]
    )
    associe = SubjectRelation(
        "Personne",
        cardinality="?*",
        constraints=[RQLConstraint("S actionnaire SOC, O actionnaire SOC")],
    )


class connait(RelationDefinition):
    subject = "CWUser"
    object = "Personne"


class Societe(EntityType):
    nom = String(fulltextindexed=True)
    web = String(maxsize=128)
    type = String(maxsize=128)  # attribute in common with Note
    tel = Int()
    fax = Int()
    rncs = String(maxsize=128)
    ad1 = String(maxsize=128)
    ad2 = String(maxsize=128)
    ad3 = String(maxsize=128)
    cp = String(maxsize=12)
    ville = String(maxsize=32)
    evaluee = SubjectRelation("Note")
    fournit = SubjectRelation(("Service", "Produit"), cardinality="1*")
    contrat_exclusif = SubjectRelation("Personne", cardinality="??")


# enough relations to cover most reledit use cases
class Project(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", "validators", "contributors"),
        "update": ("managers", "validators", "contributors"),
        "delete": ("managers", "owners"),
    }
    title = String(maxsize=32, required=True, fulltextindexed=True)
    long_desc = SubjectRelation("Blog", composite="subject", cardinality="?*")
    manager = SubjectRelation("Personne", cardinality="?*", __permissions__=rperms("O"))


class composite_card11_2ttypes(RelationDefinition):
    subject = "Project"
    object = ("File", "Blog")
    composite = "subject"
    cardinality = "?*"


class Ticket(EntityType):
    title = String(maxsize=32, required=True, fulltextindexed=True)
    concerns = SubjectRelation("Project", composite="object")
    in_version = SubjectRelation(
        "Version", composite="object", cardinality="?*", inlined=True
    )


class Version(EntityType):
    name = String(required=True)


class Filesystem(EntityType):
    name = String()


class DirectoryPermission(EntityType):
    value = String()


class parent_fs(RelationDefinition):
    name = "parent"
    subject = "Directory"
    object = "Filesystem"


class Directory(EntityType):
    name = String(required=True)
    has_permission = SubjectRelation(
        "DirectoryPermission", cardinality="*1", composite="subject"
    )


class parent_directory(RelationDefinition):
    name = "parent"
    subject = "Directory"
    object = "Directory"
    composite = "object"


class Folder(EntityType):
    name = String(required=True)
    filed_under = SubjectRelation("Folder", description=_("parent folder"))


class TreeNode(EntityType):
    name = String(required=True)
    parent = SubjectRelation("TreeNode", cardinality="?*")


class Note(EntityType):
    type = String()
    ecrit_par = SubjectRelation("Personne")


class SubNote(Note):
    __specializes_schema__ = True
    description = String()


class buddies(ComputedRelation):
    rule = "S in_group G, O in_group G"


class Ami(EntityType):
    """A Person, for which surname is not required"""

    prenom = String()
    nom = String()


class Ville(EntityType):
    insee = Int(required=True)


class Service(EntityType):
    fabrique_par = SubjectRelation("Personne", cardinality="1*")


class Produit(EntityType):
    fabrique_par = SubjectRelation("Usine", cardinality="1*", inlined=True)


class Usine(EntityType):
    lieu = String(required=True)


class evaluee(RelationDefinition):
    subject = "CWUser"
    object = "Note"


class StateFull(WorkflowableEntityType):
    name = String()


class Reference(EntityType):
    nom = String(unique=True)
    ean = String(unique=True, required=True)


class FakeFile(EntityType):
    title = String(fulltextindexed=True, maxsize=256)
    data = Bytes(required=True, fulltextindexed=True, description=_("file to upload"))
    data_format = String(
        required=True,
        maxsize=128,
        description=_(
            "MIME type of the file. Should be dynamically set at upload time."
        ),
    )
    data_encoding = String(
        maxsize=32,
        description=_(
            "encoding of the file when it applies (e.g. text). "
            "Should be dynamically set at upload time."
        ),
    )
    data_name = String(
        required=True,
        fulltextindexed=True,
        description=_("name of the file. Should be dynamically set at upload time."),
    )
    description = RichString(
        fulltextindexed=True, internationalizable=True, default_format="text/rest"
    )


class Company(EntityType):
    order = Int()
    name = String()
    description = RichString()



from cubicweb.entities import AnyEntity, fetch_config


class BlogEntry(AnyEntity):
    __regid__ = "BlogEntry"
    fetch_attrs, cw_fetch_order = fetch_config(["creation_date", "title"], order="DESC")



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""html calendar views"""


from datetime import timedelta

from logilab.common.date import todatetime
from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import adaptable
from cubicweb.utils import json_dumps, make_uid
from cubicweb_web.view import EntityView

# useful constants & functions ################################################

ONEDAY = timedelta(1)

WEEKDAYS = (
    _("monday"),
    _("tuesday"),
    _("wednesday"),
    _("thursday"),
    _("friday"),
    _("saturday"),
    _("sunday"),
)
MONTHNAMES = (
    _("january"),
    _("february"),
    _("march"),
    _("april"),
    _("may"),
    _("june"),
    _("july"),
    _("august"),
    _("september"),
    _("october"),
    _("november"),
    _("december"),
)

ICAL_EVENT = "event"
ICAL_TODO = "todo"


class ICalendarableAdapter(EntityAdapter):
    __needs_bw_compat__ = True
    __regid__ = "ICalendarable"
    __abstract__ = True

    # component type
    component = ICAL_EVENT

    @property
    def start(self):
        """return start date"""
        raise NotImplementedError

    @property
    def stop(self):
        """return stop date"""
        raise NotImplementedError


# Calendar views ##############################################################

try:
    from vobject import iCalendar

    class iCalView(EntityView):
        """A calendar view that generates a iCalendar file (RFC 5545)

        Does apply to ICalendarable compatible entities
        """

        __select__ = adaptable("ICalendarable")
        paginable = False
        content_type = "text/calendar"
        title = _("iCalendar")
        templatable = False
        __regid__ = "ical"

        def call(self):
            ical = iCalendar()
            for i in range(len(self.cw_rset.rows)):
                task = self.cw_rset.complete_entity(i, 0)
                ical_task = task.cw_adapt_to("ICalendarable")
                if ical_task.component == ICAL_TODO:
                    elt = ical.add("vtodo")
                    stop_kw = "due"
                else:
                    elt = ical.add("vevent")
                    stop_kw = "dtend"
                elt.add("uid").value = task.absolute_url()  # unique stable id
                elt.add("url").value = task.absolute_url()
                elt.add("summary").value = task.dc_title()
                elt.add("description").value = task.dc_description()
                if ical_task.start:
                    elt.add("dtstart").value = ical_task.start
                if ical_task.stop:
                    elt.add(stop_kw).value = ical_task.stop

            buff = ical.serialize()
            if not isinstance(buff, str):
                buff = str(buff, self._cw.encoding)
            self.w(buff)

except ImportError:
    pass


class hCalView(EntityView):
    """A calendar view that generates a hCalendar file

    Does apply to ICalendarable compatible entities
    """

    __regid__ = "hcal"
    __select__ = adaptable("ICalendarable")
    paginable = False
    title = _("hCalendar")
    # templatable = False

    def call(self):
        self.w('<div class="hcalendar">')
        for i in range(len(self.cw_rset.rows)):
            task = self.cw_rset.complete_entity(i, 0)
            self.w('<div class="vevent">')
            self.w('<h3 class="summary">%s</h3>', task.dc_title())
            self.w(
                '<div class="description">%s</div>',
                task.dc_description(format="text/html"),
            )
            icalendarable = task.cw_adapt_to("ICalendarable")
            if icalendarable.start:
                self.w(
                    '<abbr class="dtstart" title="%s">%s</abbr>',
                    icalendarable.start.isoformat(),
                    self._cw.format_date(icalendarable.start),
                )
            if icalendarable.stop:
                self.w(
                    '<abbr class="dtstop" title="%s">%s</abbr>',
                    icalendarable.stop.isoformat(),
                    self._cw.format_date(icalendarable.stop),
                )
            self.w("</div>")
        self.w("</div>")


class CalendarItemView(EntityView):
    __regid__ = "calendaritem"

    def cell_call(self, row, col, dates=False):
        task = self.cw_rset.complete_entity(row, 0)
        task.view("oneline", w=self.w)
        if dates:
            icalendarable = task.cw_adapt_to("ICalendarable")
            if icalendarable.start and icalendarable.stop:
                self.w(
                    "<br/> %s",
                    self._cw._("from %(date)s")
                    % {"date": self._cw.format_date(icalendarable.start)},
                )
                self.w(
                    "<br/> %s",
                    self._cw._("to %(date)s")
                    % {"date": self._cw.format_date(icalendarable.stop)},
                )
            else:
                self.w(
                    "<br/>%s",
                    self._cw.format_date(icalendarable.start or icalendarable.stop),
                )


class _TaskEntry:
    def __init__(self, task, color, index=0):
        self.task = task
        self.color = color
        self.index = index
        self.length = 1
        icalendarable = task.cw_adapt_to("ICalendarable")
        self.start = icalendarable.start
        self.stop = icalendarable.stop

    def in_working_hours(self):
        """predicate returning True is the task is in working hours"""
        if todatetime(self.start).hour > 7 and todatetime(self.stop).hour < 20:
            return True
        return False

    def is_one_day_task(self):
        return (
            self.start
            and self.stop
            and self.start.isocalendar() == self.stop.isocalendar()
        )


class CalendarView(EntityView):
    __regid__ = "calendar"
    __select__ = adaptable("ICalendarable")

    paginable = False
    title = _("calendar")

    fullcalendar_options = {
        "firstDay": 1,
        "firstHour": 8,
        "defaultView": "month",
        "editable": True,
        "header": {
            "left": "prev,next today",
            "center": "title",
            "right": "month,agendaWeek,agendaDay",
        },
    }

    def call(self, cssclass=""):
        self._cw.add_css(("fullcalendar.css", "cubicweb.calendar.css"))
        self._cw.add_js(
            (
                "jquery.ui.js",
                "fullcalendar.min.js",
                "jquery.qtip.min.js",
                "fullcalendar.locale.js",
            )
        )
        self.calendar_id = "cal" + make_uid("uid")
        self.add_onload()
        # write calendar div to load jquery fullcalendar object
        if cssclass:
            self.w('<div class="%s" id="%s"></div>', cssclass, self.calendar_id)
        else:
            self.w('<div id="%s"></div>', self.calendar_id)

    def add_onload(self):
        fullcalendar_options = self.fullcalendar_options.copy()
        fullcalendar_options["events"] = self.get_events()
        # i18n
        # js callback to add a tooltip and to put html in event's title
        js = """
        var options = $.fullCalendar.regional('%s', %s);
        options.eventRender = function(event, $element) {
          // add a tooltip for each event
          var div = '<div class="tooltip">'+ event.description+ '</div>';
          $element.append(div);
          // allow to have html tags in event's title
          $element.find('span.fc-event-title').html($element.find('span.fc-event-title').text());
        };
        $("#%s").fullCalendar(options);
        """  # "
        self._cw.add_onload(
            js % (self._cw.lang, json_dumps(fullcalendar_options), self.calendar_id)
        )

    def get_events(self):
        events = []
        for entity in self.cw_rset.entities():
            icalendarable = entity.cw_adapt_to("ICalendarable")
            if not (icalendarable.start and icalendarable.stop):
                continue
            start_date = icalendarable.start or icalendarable.stop
            event = {
                "eid": entity.eid,
                "title": entity.view("calendaritem"),
                "url": xml_escape(entity.absolute_url()),
                "className": "calevent",
                "description": entity.view("tooltip"),
            }
            event["start"] = start_date.strftime("%Y-%m-%dT%H:%M")
            event["allDay"] = True
            if icalendarable.stop:
                event["end"] = icalendarable.stop.strftime("%Y-%m-%dT%H:%M")
                event["allDay"] = False
            events.append(event)
        return events


class OneMonthCal(CalendarView):
    __regid__ = "onemonthcal"

    title = _("one month")


class OneWeekCal(CalendarView):
    __regid__ = "oneweekcal"

    title = _("one week")
    fullcalendar_options = CalendarView.fullcalendar_options.copy()
    fullcalendar_options["defaultView"] = "agendaWeek"



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""produces some Ontology Web Language schema and views

"""

from logilab.mtconverter import TransformError

from cubicweb import _
from cubicweb.predicates import none_rset, match_view
from cubicweb_web.view import StartupView, EntityView
from cubicweb_web.action import Action
from cubicweb_web.views import schema

OWL_CARD_MAP = {
    "1": '<rdf:type rdf:resource="&owl;FunctionalProperty"/>',
    "?": '<owl:maxCardinality rdf:datatype="&xsd;int">1</owl:maxCardinality>',
    "+": '<owl:minCardinality rdf:datatype="&xsd;int">1</owl:minCardinality>',
    "*": "",
}

OWL_TYPE_MAP = {
    "String": "xsd:string",
    "Bytes": "xsd:byte",
    "Password": "xsd:byte",
    "Boolean": "xsd:boolean",
    "Int": "xsd:int",
    "BigInt": "xsd:int",
    "Float": "xsd:float",
    "Decimal": "xsd:decimal",
    "Date": "xsd:date",
    "Datetime": "xsd:dateTime",
    "TZDatetime": "xsd:dateTime",
    "Time": "xsd:time",
    "TZTime": "xsd:time",
    "Interval": "xsd:duration",
}

OWL_OPENING_ROOT = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE rdf:RDF [
        <!ENTITY owl "http://www.w3.org/2002/07/owl#" >
        <!ENTITY xsd "http://www.w3.org/2001/XMLSchema#" >
]>
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema#"
    xmlns:owl="http://www.w3.org/2002/07/owl#"
    xmlns="http://logilab.org/owl/ontologies/%(appid)s#"
    xmlns:%(appid)s="http://logilab.org/owl/ontologies/%(appid)s#"
    xmlns:base="http://logilab.org/owl/ontologies/%(appid)s">

  <owl:Ontology rdf:about="">
    <rdfs:comment>
    %(appid)s Cubicweb OWL Ontology
    </rdfs:comment>
  </owl:Ontology>"""

OWL_CLOSING_ROOT = "</rdf:RDF>"


class OWLView(StartupView):
    """This view export in owl format schema database. It is the TBOX"""

    __regid__ = "owl"
    title = _("owl")
    templatable = False
    content_type = "application/xml"  # 'text/xml'

    def call(self, writeprefix=True):
        skipmeta = int(self._cw.form.get("skipmeta", True))
        if writeprefix:
            self.w(OWL_OPENING_ROOT, {"appid": self._cw.vreg.schema.name})
        self.visit_schema(skiptypes=skipmeta and schema.SKIP_TYPES or ())
        if writeprefix:
            self.w(OWL_CLOSING_ROOT)

    def should_display_rschema(self, eschema, rschema, role):
        return rschema not in self.skiptypes and (
            rschema.may_have_permission("read", self._cw, eschema, role)
        )

    def visit_schema(self, skiptypes):
        """get a layout for a whole schema"""
        self.skiptypes = skiptypes
        entities = sorted(
            eschema
            for eschema in self._cw.vreg.schema.entities()
            if not eschema.final or eschema in skiptypes
        )
        self.w("<!-- classes definition -->")
        for eschema in entities:
            self.visit_entityschema(eschema)
            self.w("<!-- property definition -->")
            self.visit_property_schema(eschema)
            self.w("<!-- datatype property -->")
            self.visit_property_object_schema(eschema)

    def visit_entityschema(self, eschema):
        """get a layout for an entity OWL schema"""
        self.w('<owl:Class rdf:ID="%s">', eschema)
        self.w("<!-- relations -->")
        for rschema, targetschemas, role in eschema.relation_definitions():
            if not self.should_display_rschema(eschema, rschema, role):
                continue
            for oeschema in targetschemas:
                card = rschema.role_relation_definition(
                    eschema, oeschema, role
                ).role_cardinality(role)
                cardtag = OWL_CARD_MAP[card]
                if cardtag:
                    self.w(
                        """<rdfs:subClassOf>
 <owl:Restriction>
  <owl:onProperty rdf:resource="#%s"/>
  %s
 </owl:Restriction>
</rdfs:subClassOf>""",
                        rschema,
                        cardtag,
                    )

        self.w("<!-- attributes -->")
        for rschema, aschema in eschema.attribute_definitions():
            if not self.should_display_rschema(eschema, rschema, "subject"):
                continue
            self.w(
                """<rdfs:subClassOf>
  <owl:Restriction>
   <owl:onProperty rdf:resource="#%s"/>
   <rdf:type rdf:resource="&owl;FunctionalProperty"/>
  </owl:Restriction>
</rdfs:subClassOf>""",
                rschema,
            )
        self.w("</owl:Class>")

    def visit_property_schema(self, eschema):
        """get a layout for property entity OWL schema"""
        for rschema, targetschemas, role in eschema.relation_definitions():
            if not self.should_display_rschema(eschema, rschema, role):
                continue
            for oeschema in targetschemas:
                self.w(
                    """<owl:ObjectProperty rdf:ID="%s">
 <rdfs:domain rdf:resource="#%s"/>
 <rdfs:range rdf:resource="#%s"/>
</owl:ObjectProperty>""",
                    rschema,
                    eschema,
                    oeschema.type,
                )

    def visit_property_object_schema(self, eschema):
        for rschema, aschema in eschema.attribute_definitions():
            if not self.should_display_rschema(eschema, rschema, "subject"):
                continue
            self.w(
                """<owl:DatatypeProperty rdf:ID="%s">
  <rdfs:domain rdf:resource="#%s"/>
  <rdfs:range rdf:resource="%s"/>
</owl:DatatypeProperty>""",
                rschema,
                eschema,
                OWL_TYPE_MAP[aschema.type],
            )


class OWLABOXView(EntityView):
    """This view represents a part of the ABOX for a given entity."""

    __regid__ = "owlabox"
    title = _("owlabox")
    templatable = False
    content_type = "application/xml"  # 'text/xml'

    def call(self):
        self.w(OWL_OPENING_ROOT, {"appid": self._cw.vreg.schema.name})
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self.w(OWL_CLOSING_ROOT)

    def cell_call(self, row, col):
        self.wview("owlaboxitem", self.cw_rset, row=row, col=col)


class OWLABOXItemView(EntityView):
    """This view represents a part of the ABOX for a given entity."""

    __regid__ = "owlaboxitem"
    templatable = False
    content_type = "application/xml"  # 'text/xml'

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        eschema = entity.e_schema
        self.w('<%s rdf:ID="%s">', eschema, entity.eid)
        self.w("<!--attributes-->")
        for rschema, aschema in eschema.attribute_definitions():
            if rschema.meta:
                continue
            rdef = rschema.relation_definition(eschema, aschema)
            if not rdef.may_have_permission("read", self._cw):
                continue
            aname = rschema.type
            if aname == "eid":
                continue
            try:
                attr = entity.printable_value(aname, format="text/plain")
                if attr:
                    self.w("<%s>%s</%s>", aname, attr, aname)
            except TransformError:
                pass
        self.w("<!--relations -->")
        for rschema, targetschemas, role in eschema.relation_definitions():
            if rschema.meta:
                continue
            for tschema in targetschemas:
                rdef = rschema.role_relation_definition(eschema, tschema, role)
                if rdef.may_have_permission("read", self._cw):
                    break
            else:
                # no read perms to any relation of this type. Skip.
                continue
            if role == "object":
                attr = "reverse_%s" % rschema.type
            else:
                attr = rschema.type
            for x in getattr(entity, attr):
                self.w("<%s>%s %s</%s>", attr, x.__regid__, x.eid, attr)
        self.w("</%s>", eschema)


class DownloadOWLSchemaAction(Action):
    __regid__ = "download_as_owl"
    __select__ = none_rset() & match_view("schema")

    category = "mainactions"
    title = _("download schema as owl")

    def url(self):
        return self._cw.build_url("view", vid="owl")



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""json export views"""


from cubicweb import _
from cubicweb.predicates import ExpectedValuePredicate, any_rset, empty_rset
from cubicweb.uilib import rest_traceback
from cubicweb.utils import json_dumps
from cubicweb_web.view import EntityView, AnyRsetView
from cubicweb_web.application import anonymized_request
from cubicweb_web.views import basecontrollers, management


class JsonpController(basecontrollers.ViewController):
    """The jsonp controller is the same as a ViewController but :

    - anonymize request (avoid CSRF attacks)
    - if ``vid`` parameter is passed, make sure it's sensible (i.e. either
      "jsonexport" or "ejsonexport")
    - if ``callback`` request parameter is passed, it's used as json padding


    Response's content-type will either be ``application/javascript`` or
    ``application/json`` depending on ``callback`` parameter presence or not.
    """

    __regid__ = "jsonp"

    def publish(self, rset=None):
        if "vid" in self._cw.form:
            vid = self._cw.form["vid"]
            if vid not in ("jsonexport", "ejsonexport"):
                self.warning(
                    "vid %s can't be used with jsonp controller, "
                    "falling back to jsonexport",
                    vid,
                )
                self._cw.form["vid"] = "jsonexport"
        else:  # if no vid is specified, use jsonexport
            self._cw.form["vid"] = "jsonexport"
        if self._cw.vreg.config["anonymize-jsonp-queries"]:
            with anonymized_request(self._cw):
                return self._get_json_data(rset)
        else:
            return self._get_json_data(rset)

    def _get_json_data(self, rset):
        json_data = super().publish(rset)
        if "callback" in self._cw.form:  # jsonp
            json_padding = self._cw.form["callback"].encode("ascii")
            # use ``application/javascript`` if ``callback`` parameter is
            # provided, keep ``application/json`` otherwise
            self._cw.set_content_type("application/javascript")
            json_data = json_padding + b"(" + json_data + b")"
        return json_data


class JsonMixIn:
    """mixin class for json views

    Handles the following optional request parameters:

    - ``_indent`` : must be an integer. If found, it is used to pretty print
      json output
    """

    templatable = False
    content_type = "application/json"
    binary = True

    def wdata(self, data):
        indent = None
        try:
            if "_indent" in self._cw.form:
                indent = int(self._cw.form["_indent"])
        except AttributeError:
            pass  # may occur with non-web connection
        # python's json.dumps escapes non-ascii characters
        self.w(json_dumps(data, indent=indent).encode("ascii"))


class JsonRsetView(JsonMixIn, AnyRsetView):
    """dumps raw result set in JSON format"""

    __regid__ = "jsonexport"
    __select__ = any_rset()  # means rset might be empty or have any shape
    title = _("json-export-view")

    def call(self):
        # XXX mimic w3c recommandations to serialize SPARQL results in json?
        #     http://www.w3.org/TR/rdf-sparql-json-res/
        self.wdata(self.cw_rset.rows)


class JsonEntityView(JsonMixIn, EntityView):
    """dumps rset entities in JSON

    The following additional metadata is added to each row :

    - ``cw_etype`` : entity type
    - ``cw_source`` : source url
    """

    __regid__ = "ejsonexport"
    __select__ = EntityView.__select__ | empty_rset()
    title = _("json-entities-export-view")

    def call(self):
        entities = []
        for entity in self.cw_rset.entities():
            serializer = entity.cw_adapt_to("ISerializable")
            entities.append(serializer.serialize())
        self.wdata(entities)


class _requested_vid(ExpectedValuePredicate):
    """predicate that checks vid parameter value

    It differs from ``match_view`` in that it doesn't expect a ``view``
    parameter to be given to ``select`` but will rather check
    ``req.form['vid']`` to match expected vid.
    """

    def __call__(self, cls, req, rset=None, **kwargs):
        return req.form.get("vid") in self.expected


class JsonErrorView(JsonMixIn, management.ErrorView):
    """custom error view selected when client asks for a json view

    The returned json object will contain err / traceback informations.
    """

    __select__ = management.ErrorView.__select__ & _requested_vid(
        "jsonexport", "ejsonexport"
    )

    def call(self):
        errmsg, exclass, excinfo = self._excinfo()
        self.wdata(
            {
                "errmsg": errmsg,
                "exclass": exclass,
                "traceback": rest_traceback(excinfo, errmsg),
            }
        )



from urllib.parse import urljoin
from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import PyramidWebCWTC
from cubicweb_web.view import View
from cubicweb_web._exceptions import Redirect


class Redirector(View):
    __regid__ = "redirector"

    def call(self, rset=None):
        self._cw.set_header("Cache-Control", "no-cache")
        raise Redirect("http://example.org")


class PyramidCoreTest(PyramidWebCWTC):
    anonymous_allowed = True

    def test_cw_to_pyramid_copy_headers_on_redirect(self):
        self.vreg.register(Redirector)
        try:
            res = self.webapp.get(
                urljoin(BASE_URL, "?vid=redirector"), expect_errors=True
            )
            self.assertEqual(res.status_int, 303)
            self.assertEqual(res.headers["Cache-Control"], "no-cache")
        finally:
            self.vreg.unregister(Redirector)


if __name__ == "__main__":
    from unittest import main

    main()
